"""이 모듈은 날짜와 관련된 유틸리티 함수를 제공합니다.

- `to_string` 함수는 datetime 객체를 문자열로 변환합니다.
- `is_valid` 함수는 주어진 날짜가 특정 카테고리에 대해 유효한지 확인합니다.
"""

from datetime import datetime

import maplestory.utils.kst as kst
from maplestory.enums import QueryableDate

DATE_FORMAT = "%Y-%m-%d"


def to_string(date: datetime | None, format=DATE_FORMAT) -> str | None:
    """datetime 객체를 주어진 포맷의 문자열로 변환합니다. 만약 date가 None이면 None을 반환합니다.

    Args:
        date (datetime | None): 변환할 datetime 객체입니다.
        format (str, optional): 날짜를 문자열로 변환할 때 사용할 포맷입니다. 기본값은 "%Y-%m-%d"입니다.

    Returns:
        str | None: datetime 객체를 문자열로 변환한 결과입니다. date가 None이면 None을 반환합니다.
    """
    return None if date is None else date.strftime(format)


def is_valid(date: kst.AwareDatetime, category: QueryableDate) -> None:
    """주어진 날짜가 특정 카테고리에 대해 유효한지 확인합니다. 만약 날짜가 카테고리의 조회 가능한 날짜보다 이전이면 ValueError를 발생시킵니다.

    Args:
        date (kst.AwareDatetime): 확인할 날짜
        category (MinimumDate): 확인할 카테고리

    Raises:
        ValueError: 날짜가 카테고리의 조회 가능한 날짜보다 이전인 경우 발생합니다.
    """

    kst.validate(date)

    if date < category.value:
        queryable_date = category.value.strftime(DATE_FORMAT)
        raise ValueError(
            f"{category.name}은 {queryable_date}부터 데이터를 조회할 수 있습니다."
        )



from datetime import datetime

from pydantic import BaseModel, Field


class SetEffectOptionInfo(BaseModel):
    """세트 효과 옵션 정보

    Attributes:
        set_count (int): 세트 효과 레벨 (장비 수)
        set_option (str): 적용 중인 세트 효과
    """

    set_count: int
    set_option: str


class SetEffectInfo(BaseModel):
    """세트 효과 정보

    Attributes:
        set_name (str): 세트 효과 명
        total_set_count (int): 세트 개수 (럭키 아이템 포함)
        set_effect_info (list[SetEffectOptionInfo]): 세트 효과 옵션
    """

    set_name: str
    total_set_count: int
    set_effect_info: list[SetEffectOptionInfo]


class SetEffect(BaseModel):
    """세트 효과 정보

    Attributes:
        date (datetime): 조회 기준일
        set_effect (list[SetEffectInfo]): 세트 효과 정보

    Examples:
        {
            "date": "2024-01-25T00:00+09:00",
            "set_effect": [
                {
                    "set_name": "루타비스 세트(전사)",
                    "total_set_count": 2,
                    "set_effect_info": [
                        {
                            "set_count": 2,
                            "set_option": "STR : +20, DEX : +20, 최대 HP : +1000, 최대 MP : +1000"
                        }
                    ]
                },
                {
                    "set_name": "마이스터 세트",
                    "total_set_count": 1,
                    "set_effect_info": []
                },
                {
                    "set_name": "보스 장신구 세트",
                    "total_set_count": 6,
                    "set_effect_info": [
                        {
                            "set_count": 3,
                            "set_option": "올스탯 : +10, 최대 HP : +5%, 최대 MP : +5%, 공격력 : +5, 마력 : +5"
                        },
                        {
                            "set_count": 5,
                            "set_option": "올스탯 : +10, 최대 HP : +5%, 최대 MP : +5%, 공격력 : +5, 마력 : +5"
                        }
                    ]
                },
                {
                    "set_name": "앱솔랩스 세트(전사)",
                    "total_set_count": 4,
                    "set_effect_info": [
                        {
                            "set_count": 2,
                            "set_option": "최대 HP : +1500, 최대 MP : +1500, 공격력 : +20, 마력 : +20, 보스 몬스터 공격 시 데미지 : +10%"
                        },
                        {
                            "set_count": 3,
                            "set_option": "올스탯 : +30, 공격력 : +20, 마력 : +20, 보스 몬스터 공격 시 데미지 : +10%"
                        },
                        {
                            "set_count": 4,
                            "set_option": "공격력 : +25, 마력 : +25, 방어력 : +200, 몬스터 방어율 무시 : +10%"
                        }
                    ]
                },
                {
                    "set_name": "아케인셰이드 세트(전사)",
                    "total_set_count": 1,
                    "set_effect_info": []
                }
            ]
        }
    """

    date: datetime = Field(repr=False)
    set_effect: list[SetEffectInfo]



from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field, field_validator

from maplestory.models.types import PageCursor


# FIXME: plus_value가 현재 str인데, int/float으로 변경하기. 현재 'null'밖에 보지 못함.
# TODO: success_rate, event_range가 리스트인 경우 별도로 분리하기
class StarforceHistoryEvent(BaseModel):
    """진행 중인 스타포스 강화 이벤트 정보

    Attributes:
        success_rate: 이벤트 성공 확률
        cost_discount_rate: 이벤트 비용 할인율
        plus_value: 이벤트 강화 수치 가중값
        starforce_event_range: 이벤트 적용 강화 시도 가능한 n성 범위

    Examples:
        {
            "success_rate": "null",             -> None
            "cost_discount_rate": "30",         -> 0.3
            "plus_value": "null",               -> None
            "starforce_event_range": "0~24"
        },
        {
            "success_rate": "100,100,100",
            "cost_discount_rate": "null",       -> None
            "plus_value": "null",               -> None
            "starforce_event_range": "5,10,15"
        }
    """

    success_rate: str | None
    cost_discount_rate: float | None
    plus_value: str | None
    starforce_event_range: str

    @field_validator("success_rate", mode="before")
    @classmethod
    def change_success_rate(cls, v: Any) -> str | None:
        return None if v == "null" else v

    @field_validator("cost_discount_rate", mode="before")
    @classmethod
    def change_cost_discount_rate(cls, v: Any) -> float | None:
        return None if v == "null" else int(v) / 100

    @field_validator("plus_value", mode="before")
    @classmethod
    def change_plus_value(cls, v: Any) -> Any:
        return None if v == "null" else v


class StarforceHistoryInfo(BaseModel):
    """스타포스 히스토리 정보

    Attributes:
        id: 스타포스 히스토리 식별자
        item_upgrade_result: 강화 시도 결과
        before_starforce_count: 강화 시도 전 스타포스 수치
        after_starforce_count: 강화 시도 후 스타포스 수치
        starcatch_result: 스타 캐치
        superior_item_flag: 슈페리얼 장비
        destroy_defense: 파괴 방지
        chance_time: 찬스 타임
        event_field_flag: 파괴 방지 필드 이벤트
        upgrade_item: 사용 주문서 명
        protect_shield: 프로텍트 실드
        bonus_stat_upgrade: 보너스 스탯 부여 아이템 여부
        character_name: 캐릭터 명
        world_name: 월드 명
        target_item: 대상 장비 아이템 명
        date_create: 강화 일시 (KST) example: 2023-12-27T17:28+09:00
        starforce_event_list: 진행 중인 스타포스 강화 이벤트 정보
    """

    id: str  # Len(96, 96)
    item_upgrade_result: str
    before_starforce_count: int
    after_starforce_count: int
    starcatch_result: str
    superior_item_flag: str
    destroy_defense: str = Field(alias="destroy_defence")
    chance_time: str
    event_field_flag: str
    upgrade_item: str
    protect_shield: str
    bonus_stat_upgrade: str
    character_name: str
    world_name: str
    target_item: str
    date_create: datetime
    starforce_event_list: list[StarforceHistoryEvent]


class StarforceHistory(BaseModel):
    """스타포스 히스토리

    Attributes:
        count: 결과 건 수
        next_cursor: 페이징 처리를 위한 cursor
        starforce_history: 스타포스 히스토리
    """

    count: int
    next_cursor: PageCursor | None
    starforce_history: list[StarforceHistoryInfo]



"""이 모듈은 Pydantic의 BaseModel을 상속받은 클래스에서 특정 조건을 만족하는 필드를 `__repr__` 메소드의 결과에서 제외하는 기능을 제공합니다.

- `HideNoneRepresentation` 믹스인 클래스는 `None` 값을 가진 필드를 제외합니다.
- `HideZeroStatRepresentation` 믹스인 클래스는 "STR", "INT", "DEX", "LUK" 키에 대응하는 값이 0인 필드를 제외합니다.
"""

from datetime import datetime
from typing import Any
from zoneinfo import ZoneInfo

from pydantic import BaseModel
from pydantic_core import TzInfo

from maplestory.utils.kst import KST_TZ_CONSTRAINT


class HideNoneRepresentation:
    """`None` 값을 가진 필드를 `__repr__` 메서드의 결과에서 제외하는 믹스인 클래스입니다."""

    def __repr_args__(self: BaseModel):
        """BaseModel의 `__repr__` 메서드에서 사용하는 필드와 값의 리스트를 반환합니다.

        이 메서드는 `None` 값을 가진 필드를 제외합니다.

        Returns:
            list: 필드 이름과 값의 튜플을 원소로 가지는 리스트입니다. `None` 값을 가진 필드는 제외됩니다.
        """
        return [
            (key, value) for key, value in self.__dict__.items() if value is not None
        ]


class HideZeroStatRepresentation:
    """스탯 값이 0인 경우를 `__repr__` 메서드의 결과에서 제외하는 믹스인 클래스입니다."""

    STAT_KEYS = ["STR", "INT", "DEX", "LUK", "HP"]

    def __repr_args__(self: BaseModel):
        """BaseModel의 `__repr__` 메소드에서 사용하는 필드와 값의 리스트를 반환하는데, 스탯 값이 0인 필드를 제외합니다.

        Returns:
            list: 스탯 값이 0인 필드가 제외된 필드 이름과 값의 튜플을 원소로 가지는 리스트입니다.
        """

        return [
            (key, value)
            for key, value in self.__dict__.items()
            if key not in self.STAT_KEYS or value != 0
        ]


class DatetimeRepresentation:
    def convert_datetime_as_sstr(self, value: Any):
        if not isinstance(value, datetime):
            return value

        tz = ""
        if value.tzinfo in [ZoneInfo(key="Asia/Seoul"), TzInfo(KST_TZ_CONSTRAINT)]:
            tz = " KST"

        if value.hour or value.minute or value.second:
            return value.strftime(f"%Y-%m-%d %H:%M:%S{tz}")

        return value.strftime(f"%Y-%m-%d{tz}")

    def __repr_args__(self: BaseModel):
        for k, v in self.__dict__.items():
            field = self.model_fields.get(k)
            if field and field.repr:
                yield k, self.convert_datetime_as_sstr(v)

        try:
            pydantic_extra = object.__getattribute__(self, "__pydantic_extra__")
        except AttributeError:
            pydantic_extra = None

        if pydantic_extra is not None:
            yield from ((k, v) for k, v in pydantic_extra.items())
        yield from (
            (k, getattr(self, k))
            for k, v in self.model_computed_fields.items()
            if v.repr
        )



from __future__ import annotations

from enum import Enum

from httpx import HTTPStatusError, Response
from pydantic import BaseModel, Field, computed_field


class APIErrorCode(str, Enum):
    """API 에러 코드를 나타내는 Enum 클래스입니다. 각 코드는 특정 API 에러 상황을 나타냅니다.

    References:
        https://openapi.nexon.com/guide/request-api/#error-code-table
    """

    OPENAPI00001 = "서버 내부 오류"
    OPENAPI00002 = "권한이 없는 경우"
    OPENAPI00003 = "유효하지 않은 식별자"
    OPENAPI00004 = "파라미터 누락 또는 유효하지 않음"
    OPENAPI00005 = "유효하지 않은 API KEY"
    OPENAPI00006 = "유효하지 않은 게임 또는 API PATH"
    OPENAPI00007 = "API 호출량 초과"
    OPENAPI00009 = "데이터 준비 중"

    @classmethod
    def from_code(cls, code: str) -> APIErrorCode:
        """
        문자열 코드를 해당 Enum 값으로 변환합니다.

        Args:
            code (str): 변환할 문자열 코드입니다.

        Returns:
            APIErrorCode: 해당하는 Enum 값입니다.
        """
        return cls._member_map_[code]


class ErrorMessage(BaseModel):
    """에러 메시지를 나타내는 클래스입니다.

    각 에러 메시지는 특정 에러 코드(`code`)와 그에 대한 설명(`message`)을 포함합니다.

    Attributes:
        code (str): 에러 코드 (예: OPENAPI00001)
        message (str): 에러 설명

    Examples:
        >>> error = ErrorMessage(name="OPENAPI00004", message="Please input valid parameter")
        >>> error.code
        'OPENAPI00004'
        >>> error.message
        'Please input valid parameter'
    """

    code: str = Field(alias="name")
    message: str


class APIError(HTTPStatusError):
    """API 에러를 나타내는 클래스입니다.

    각 API 에러는 에러 코드(`code`), 에러 메시지(`message`), 그리고 HTTP 상태 코드(`status`)를 포함합니다.

    Attributes:
        code (str): 에러 코드
        message (str): 에러 메시지
        status (int): HTTP 상태 코드
        description (str): 에러 코드에 대한 설명
    """

    code: str
    message: str
    status: int

    def __init__(
        self,
        error: ErrorMessage,
        response: Response,
    ):
        super().__init__(
            f"{error.message} ({error.code})",
            request=response.request,
            response=response,
        )
        self.code = error.code
        self.message = error.message
        self.status = response.status_code

    @computed_field
    @property
    def description(self) -> str:
        """에러 코드에 대한 설명을 반환합니다."""
        return APIErrorCode.from_code(self.code).value



"""
이 모듈은 메이플스토리 게임의 유니온 정보를 가져오고 처리하는 함수를 제공합니다.
This module provides functions to fetch and handle Union information in the MapleStory game.

이 모듈에는 주어진 캐릭터 식별자(ocid)와 참조 날짜에 대한 유니온 레벨과 유니온 등급 정보를 가져오는 `get_union_info` 함수가 포함되어 있습니다.
It includes a function `get_union_info` which fetches the Union level and Union rank information 
for a given character identifier (ocid) and reference date.

이 모듈은 `maplestory.models.union` 패키지의 `Union`과 `UnionRaider` 모델과 `maplestory.utils.date`와 `maplestory.utils.network` 패키지의 유틸리티 함수를 사용합니다.
The module uses the `Union` and `UnionRaider` models from the `maplestory.models.union` package 
and utility functions from the `maplestory.utils.date` and `maplestory.utils.network` packages.
"""
from __future__ import annotations

import re
from datetime import datetime
from typing import Any, Generator

from pydantic import BaseModel, Field, RootModel, computed_field, field_validator

from maplestory.apis.union import (
    get_union_artifact_info_by_ocid,
    get_union_info_by_ocid,
    get_union_raider_info_by_ocid,
)
from maplestory.models.union import UnionArtifact, UnionInfo, UnionRaider
from maplestory.models.union.artifact import UnionArtifactEffect
from maplestory.services.character import get_character_id
from maplestory.utils.kst import yesterday


class Union(BaseModel):
    character_name: str
    date: datetime = yesterday()

    @computed_field(repr=False)
    @property
    def union_info(self) -> UnionInfo:
        return get_union_info(self.character_name, self.date)

    @computed_field
    @property
    def level(self) -> int:
        return self.union_info.level

    @computed_field
    @property
    def grade(self) -> str:
        return self.union_info.grade

    @computed_field(repr=False)
    @property
    def raider_info(self) -> UnionRaider:
        return get_union_raider_info(self.character_name, self.date)

    @computed_field
    @property
    def raider_stats(self) -> UnionStats:
        stats = UnionStats.parse_list(self.raider_info.raider_stats)
        return stats.summarize(split_multiple=True)

    @property
    def 공격대원효과(self) -> UnionStats:
        return self.raider_stats

    @computed_field
    @property
    def occupied_stats(self) -> UnionStats:
        stats = UnionStats.parse_list(self.raider_info.occupied_stats)
        return stats.summarize(split_multiple=True)

    @property
    def 공격대점령효과(self) -> UnionStats:
        return self.occupied_stats

    @computed_field(repr=False)
    @property
    def artifact(self) -> UnionArtifact:
        return get_union_artifact_info(self.character_name, self.date)

    @computed_field
    @property
    def artifact_effects(self) -> list[UnionArtifactEffect]:
        return self.artifact.effects

    @property
    def 아티팩트효과(self) -> list[UnionArtifactEffect]:
        return self.artifact_effects


def get_union_info(
    character_name: str,
    date: datetime = yesterday(),
) -> UnionInfo:
    """
    유니온 레벨 및 유니온 등급 정보를 조회합니다.
    Fetches the Union level and Union rank information.

    Args:
        character_name : 캐릭터의 이름.
                         The name of the character.
        date: 조회 기준일 (KST).
                        Reference date for the query (KST).

    Returns:
        Union: 유니온의 레벨 및 등급 정보.
               The level and rank information of the Union.
    """

    character_ocid = get_character_id(character_name)
    return get_union_info_by_ocid(character_ocid, date)


def get_union_raider_info(
    character_name: str,
    date: datetime = yesterday(),
) -> UnionRaider:
    """
    유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등 상세 정보를 조회합니다.
    Fetches the detailed information of the effect of the attack members placed in the Union and the effect of the attack occupation.

    Args:
        character_name : 캐릭터의 이름.
                         The name of the character.
        date: 조회 기준일 (KST).
                        Reference date for the query (KST).

    Returns:
        UnionRaider: 유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등의 상세 정보.
                     The detailed information of the effect of the attack members placed in the Union and the effect of the attack occupation.
    """

    character_ocid = get_character_id(character_name)
    return get_union_raider_info_by_ocid(character_ocid, date)


def summarize_union_stat(stat: str) -> dict[str, str]:
    if m := re.search(
        r"(?P<name>.+) (?P<value>\d+(\.\d+)?)(?P<percent>%?) (?P<updown>.+)", stat
    ):
        return m.groupdict()

    raise ValueError("Invalid stat format.")


class UnionStat(BaseModel):
    stat: str

    @field_validator("stat", mode="before")
    @classmethod
    def unify_format(cls, stat: str) -> str:
        # '공격력 20, 마력 20 증가' -> '공격력/마력 20 증가'
        pattern = r"(\w+) (\d+), (\w+) (\d+) (\w+)"
        if m := re.match(pattern, stat):
            if m[2] == m[4]:
                replacement = r"\1/\3 \2 \5"
                return re.sub(pattern, replacement, stat)

        # '몬스터 방어율 무시 20% 증가' -> '방어율 무시 20% 증가'
        if "몬스터 방어율 무시" in stat:
            stat = stat.replace("몬스터 ", "")

        return stat

    @computed_field(repr=False)
    @property
    def parsed_stat(self) -> dict[str, str]:
        return summarize_union_stat(self.stat)

    @computed_field(repr=False)
    @property
    def name(self) -> str:
        return self.parsed_stat.get("name")

    @property
    def is_multi_name(self) -> bool:
        return "," in self.name or "/" in self.name

    @property
    def multi_name(self) -> str | None:
        if "," in self.name:
            # 'STR, DEX, LUK 40 증가'
            return [s.strip() for s in self.name.split(",")]
        elif "/" in self.name:
            # '공격력/마력 20 증가'
            return [s.strip() for s in self.name.split("/")]
        return None

    @computed_field(repr=False)
    @property
    def percent(self) -> str:
        return self.parsed_stat.get("percent")

    @computed_field(repr=False)
    @property
    def is_percent(self) -> bool:
        return self.percent == "%"

    @computed_field(repr=False)
    @property
    def value(self) -> int | float:
        try:
            return int(self.parsed_stat.get("value"))
        except ValueError as _:
            return float(self.parsed_stat.get("value"))

    @computed_field(repr=False)
    @property
    def updown(self) -> str:
        return self.parsed_stat.get("updown")

    def __add__(self, other: UnionStat) -> UnionStat:
        if self.name != other.name:
            raise ValueError("Cannot add different stats.")
        if self.updown != other.updown:
            raise ValueError("Cannot add stats with different updown.")
        if self.is_percent != other.is_percent:
            raise ValueError(
                "Cannot add stats with different types (percent vs number)."
            )

        new_value = self.value + other.value
        if self.is_percent:
            new_value = f"{new_value}%"
        return UnionStat(stat=f"{self.name} {new_value} {self.updown}")

    def __radd__(self, other: UnionStat) -> UnionStat:
        if other == 0:
            return self
        return self.__add__(other) if isinstance(other, UnionStat) else NotImplemented


class UnionStats(RootModel):
    root: list[UnionStat]

    def __repr__(self) -> str:
        return f"UnionStats({self.root!r})"

    def __rich_repr__(self) -> Generator:
        yield self.root

    def model_post_init(self, __context: Any) -> None:
        """
        모델 생성 후 초기화를 수행합니다. 스탯의 이름을 기준으로 root를 정렬합니다.

        Args:
            __context (Any): 초기화 후의 컨텍스트.
        """
        """
        Initializes the model post creation. Sorts the root based on the name of the stat.

        Args:
            __context (Any): The context for post initialization.
        """

        self.root = sorted(self.root, key=lambda x: x.name)

    def group_stats(self, split_multiple: bool = False) -> dict[str, list[UnionStat]]:
        """
        스탯의 이름과 퍼센트 여부를 기준으로 스탯을 그룹화합니다.
        split_multiple이 True인 경우, 다중 이름 스탯을 개별 스탯으로 분리합니다.

        Args:
            split_multiple (bool): 다중 이름 스탯을 분리할지 여부. 기본값은 False입니다.

        Returns:
            Dict[str, List[UnionStat]]: 그룹화된 스탯.
        """
        """
        Groups the stats based on the name and whether it is a percentage. 
        If split_multiple is True, splits the multi-name stats into individual stats.

        Args:
            split_multiple (bool): Whether to split multi-name stats. Default is False.

        Returns:
            Dict[str, List[UnionStat]]: The grouped stats.
        """
        stats = []
        if split_multiple:
            for stat in self.root:
                if stat.is_multi_name:
                    stats.extend(
                        [
                            UnionStat(
                                stat=f"{name} {stat.value}{stat.percent} {stat.updown}"
                            )
                            for name in stat.multi_name
                        ]
                    )
                else:
                    stats.append(stat)
        else:
            stats = self.root

        return {
            (stat.name, stat.is_percent): [
                s
                for s in stats
                if s.name == stat.name and s.is_percent == stat.is_percent
            ]
            for stat in stats
        }

    def group_with_split_multiple(self) -> dict[str, list[UnionStat]]:
        """
        split_multiple을 True로 설정하여 스탯을 그룹화합니다.

        Returns:
            Dict[str, List[UnionStat]]: 그룹화된 스탯.
        """
        """
        Groups the stats with split_multiple set to True.

        Returns:
            Dict[str, List[UnionStat]]: The grouped stats.
        """
        return self.group_stats(split_multiple=True)

    @classmethod
    def parse_list(cls, data: list[str]) -> UnionStats:
        """
        스탯의 리스트를 파싱하여 UnionStats 객체를 생성합니다.

        Args:
            data (List[str]): 스탯의 리스트.

        Returns:
            UnionStats: 생성된 UnionStats 객체.
        """
        """
        Parses a list of stats and creates a UnionStats object.

        Args:
            data (List[str]): The list of stats.

        Returns:
            UnionStats: The created UnionStats object.
        """
        return cls(root=[UnionStat(stat=stat) for stat in data])

    def summarize(self, split_multiple: bool = False) -> UnionStats:
        """
        스탯을 요약합니다. '방어율 무시'는 합이 아닌 곱으로 처리되므로 별도로 처리합니다.

        Args:
            split_multiple (bool): 다중 이름 스탯을 분리할지 여부. 기본값은 False입니다.

        Returns:
            UnionStats: 요약된 스탯.
        """
        """
        Summarizes the stats. '방어율 무시' (Ignore Defense Rate) is handled separately 
        as it is not summed but multiplied.

        Args:
            split_multiple (bool): Whether to split multi-name stats. Default is False.

        Returns:
            UnionStats: The summarized stats.
        """
        combined_stats = []
        grouped_stats = self.group_stats(split_multiple)

        for (stat_name, _), stats in grouped_stats.items():
            if stat_name == "방어율 무시":
                # '방어율 무시' (Ignore Defense Rate) is multiplied, not summed
                # 방어율 무시는 합적용이 아니라 곱적용이라 별도로 처리
                combined_stats.extend(stats)
            else:
                combined_stats.append(sum(stats))

        return UnionStats(root=combined_stats)


def get_union_artifact_info(
    character_name: str,
    date: datetime = yesterday(),
) -> UnionArtifact:
    """
    유니온 아티팩트 정보를 조회합니다.
    Fetches the Union Artifact information.

    Args:
        character_name : 캐릭터의 이름.
                         The name of the character.
        date: 조회 기준일 (KST).
                        Reference date for the query (KST).

    Returns:
        UnionArtifact: 유니온 아티팩트 정보.
                       The information of the Union Artifact.
    """

    character_ocid = get_character_id(character_name)
    return get_union_artifact_info_by_ocid(character_ocid, date)



import httpx
from httpx import Response
from tenacity import (
    retry,
    retry_if_exception_message,
    retry_if_exception_type,
    wait_exponential,
)

from maplestory.config import Config
from maplestory.error import APIError, ErrorMessage


# retry=retry_if_exception_type(APIError),
# retry=(retry_if_exception_type(APIError) & retry_if_exception_message(match="^.*Please try again later.*$"))
@retry(
    retry=retry_if_exception_message(match="^.*Please try again later.*$"),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
def fetch(url_path: str, query_params: dict | None = None) -> dict:
    """
    Fetches data from the API.

    Args:
        url_path : The API path.
        query_params (dict): The query parameters.

    Returns:
        dict: The response data.

    Raises:
        APIError: If an API error occurs.
    """
    config = Config()

    resp: Response = httpx.get(
        f"{config.url}{url_path}",
        params=query_params,
        headers={
            "x-nxopen-api-key": config.key,
            "Accept": "application/json",
        },
        timeout=config.timeout,
    )

    data = resp.json()
    # try:
    #     ErrorResponse.model_validate(data)
    # except ValidationError as err:

    if error := data.get("error"):
        em = ErrorMessage.model_validate(error)
        err = APIError(em, resp)
        print("--------------------------------")
        print(repr(err))
        print(repr(err.code))
        print(repr(err.status))
        print(f"request: {resp.request}")
        print(f"request.headers: {resp.request.headers}")
        print("--------------------------------")
        raise err

    return data


# Please input valid parameter
# OPENAPI00004
# 400
# Headers([('content-type', 'application/json;charset=UTF-8'), ('transfer-encoding', 'chunked'), ('connection', 'keep-alive'), ('date', 'Fri, 26 Jan 2024 07:12:40 GMT'), ('vary', 'Origin,Access-Control-Request-Method,Access-Control-Request-Headers'), ('x-content-type-options', 'nosniff'), ('x-xss-protection', '0'), ('cache-control', 'no-cache, no-store, max-age=0, must-revalidate'), ('pragma', 'no-cache'), ('expires', '0'), ('strict-transport-security', 'max-age=31536000 ; includeSubDomains'), ('referrer-policy', 'no-referrer'), ('x-envoy-upstream-service-time', '31'), ('inface-wasm-filter', '1.5.5'), ('server', 'inface'), ('x-request-id', 'ZAdjsYx19504vXRwAlmyRFManOKFp2JBznT-Gz9pBRzR3YS_Dn9KDQ=='), ('x-cache', 'Error from cloudfront'), ('via', '1.1 bca5030f468e569e9304a18600bffc26.cloudfront.net (CloudFront)'), ('x-amz-cf-pop', 'ICN54-C2'), ('x-amz-cf-id', 'ZAdjsYx19504vXRwAlmyRFManOKFp2JBznT-Gz9pBRzR3YS_Dn9KDQ=='), ('vary', 'Origin')])
# {"error":{"name":"OPENAPI00004","message":"Please input valid parameter"}}
# Please try again later (OPENAPI00007)



"""
이 모듈은 메이플스토리 게임의 유니온 정보를 가져오고 처리하는 함수를 제공합니다.
This module provides functions to fetch and handle Union information in the MapleStory game.

이 모듈에는 주어진 캐릭터 식별자(ocid)와 참조 날짜에 대한 유니온 레벨과 유니온 등급 정보를 가져오는 `get_union_info` 함수가 포함되어 있습니다.
It includes a function `get_union_info` which fetches the Union level and Union rank information 
for a given character identifier (ocid) and reference date.

이 모듈은 `maplestory.models.union` 패키지의 `Union`과 `UnionRaider` 모델과 `maplestory.utils.date`와 `maplestory.utils.network` 패키지의 유틸리티 함수를 사용합니다.
The module uses the `Union` and `UnionRaider` models from the `maplestory.models.union` package 
and utility functions from the `maplestory.utils.date` and `maplestory.utils.network` packages.
"""

from datetime import datetime

import maplestory.utils.date as dates
import maplestory.utils.kst as kst
from maplestory.models.union import UnionArtifact, UnionInfo, UnionRaider
from maplestory.utils.network import fetch


def get_union_info_by_ocid(
    character_ocid: str,
    date: datetime = kst.yesterday(),
) -> UnionInfo:
    """
    유니온 레벨 및 유니온 등급 정보를 조회합니다.
    Fetches the Union level and Union rank information.

    Args:
        character_ocid: 캐릭터의 식별자(ocid).
                        The identifier (ocid) of the character.
        date: 조회 기준일 (KST).
                        Reference date for the query (KST).

    Returns:
        Union: 유니온의 레벨 및 등급 정보.
               The level and rank information of the Union.
    """

    path = "/maplestory/v1/user/union"
    query = {
        "ocid": character_ocid,
        "date": dates.to_string(date),
    }
    response = fetch(path, query)

    return UnionInfo.model_validate(response)


def get_union_raider_info_by_ocid(
    character_ocid: str,
    date: datetime = kst.yesterday(),
) -> UnionRaider:
    """
    유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등 상세 정보를 조회합니다.
    Fetches the detailed information of the effect of the attack members placed in the Union and the effect of the attack occupation.

    Args:
        character_ocid: 캐릭터의 식별자(ocid).
                        The identifier (ocid) of the character.
        date: 조회 기준일 (KST).
                        Reference date for the query (KST).

    Returns:
        UnionRaider: 유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등의 상세 정보.
                     The detailed information of the effect of the attack members placed in the Union and the effect of the attack occupation.
    """

    path = "/maplestory/v1/user/union-raider"
    query = {
        "ocid": character_ocid,
        "date": dates.to_string(date),
    }
    response = fetch(path, query)

    return UnionRaider.model_validate(response)


def get_union_artifact_info_by_ocid(
    character_ocid: str,
    date: datetime = kst.yesterday(),
) -> UnionArtifact:
    """
    유니온 아티팩트 정보를 조회합니다.
    Fetches the Union Artifact information.

    Args:
        character_ocid: 캐릭터의 식별자(ocid).
                        The identifier (ocid) of the character.
        date: 조회 기준일 (KST).
                        Reference date for the query (KST).

    Returns:
        UnionArtifact: 유니온 아티팩트 정보.
                       The information of the Union Artifact.
    """

    path = "/maplestory/v1/user/union-artifact"
    query = {
        "ocid": character_ocid,
        "date": dates.to_string(date),
    }
    response = fetch(path, query)

    return UnionArtifact.model_validate(response)



from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, model_validator

import maplestory.utils.kst as kst
from maplestory.models.history import Account, CubeHistory, StarforceHistory
from maplestory.utils.network import fetch


def get_account_id() -> str:
    """계정 식별자(ouid) 조회

    Returns:
        str: _description_
    """

    path = "/maplestory/v1/ouid"
    response = fetch(path)

    return Account.model_validate(response).ouid


def validate_date_and_cursor(
    date: Optional[datetime] = None, cursor: Optional[str] = None
) -> None:
    """
    date와 cursor의 유효성을 확인하는 함수입니다.

    Args:
        date (Optional[datetime]): 날짜 정보. 기본값은 None입니다.
        cursor (Optional[str]): 커서 정보. 기본값은 None입니다.

    Raises:
        ValueError: date와 cursor가 동시에 제공되는 경우 또는 둘 다 제공되지 않는 경우 발생합니다.
    """

    if date is not None and cursor is not None:
        raise ValueError("date and cursor cannot be used together")

    if date is None and cursor is None:
        raise ValueError("date or cursor must be provided")


class CubeHistory:
    # history: list[CubeHistory]

    def fetch(self, date: datetime = kst.yesterday()):
        all_history = []
        history = get_cube_usage_history(reference_data=date)
        all_history.extend(history.cube_history)
        while cursor := history.next_cursor:
            history = get_cube_usage_history(cursor=cursor)
            all_history.extend(history.cube_history)

        return all_history

    def fetch_from_date(self, from_date: datetime):
        all_history = []
        the_date = from_date
        while the_date < kst.kst_today():
            history = get_cube_usage_history(reference_data=the_date)
            all_history.extend(history.cube_history)
            the_date += datetime.timedelta(days=1)

        return all_history

    def fetch_from_to(self, from_date: datetime, to_date: datetime):
        all_history = []
        the_date = from_date
        while the_date <= to_date:
            history = get_cube_usage_history(reference_data=the_date)
            all_history.extend(history.cube_history)
            the_date += datetime.timedelta(days=1)

        return all_history


def get_cube_usage_history(
    result_count: int = 1000,
    date: datetime = kst.yesterday(),
    page_cursor: str | None = None,
) -> CubeHistory:
    """
    큐브 사용 결과를 조회합니다.
    Fetches the usage result of the cube.

    Args:
        result_count : 한번에 가져오려는 결과의 갯수. The number of results to fetch at once.
        date : 조회 기준일(KST). Reference date for the query (KST).
        page_cursor : 페이징 처리를 위한 cursor. Cursor for paging.

    Returns:
        CubeHistory: 큐브 사용 결과. The usage result of the cube.
    """

    validate_date_and_cursor(date, page_cursor)

    path = "/maplestory/v1/history/cube"
    query = {
        "count": result_count,
        "date": kst.validate(date),
        "cursor": page_cursor,
    }
    response = fetch(path, query)

    return CubeHistory.model_validate(response)


# TODO: date랑 page_cursor가 같이 들어오면 에러를 발생시키도록
class StarforceHistoryParameter(BaseModel):
    result_count: int = 1000
    date: datetime | None = None  # = date.kst_yesterday()
    page_cursor: str | None = None
    # 아니면 page_cursor가 None이 아니면 date를 None으로 변경하던가

    @model_validator(mode="before")
    @classmethod
    def validate(cls, data: Any) -> Any:
        if data.get("date") and data.get("page_cursor"):
            raise ValueError("page_cursor cannot be used with date")
        return data


def get_starforce_history(
    result_count: int = 1000,
    date: datetime | None = kst.yesterday(),
    page_cursor: str | None = None,
) -> StarforceHistory:
    """
    스타포스 강화 결과를 조회합니다.
    Fetches the starforce enhancement results.

    Args:
        result_count (int): 한번에 가져오려는 결과의 갯수(최소 10, 최대 1000)
                            The number of results to fetch at once (minimum 10, maximum 1000)
        date (datetime, optional): 조회 기준일(KST) (cursor가 없는 경우 필수이며 cursor와 함께 사용 불가)
                                             Query base date (KST) (required if cursor is not present and cannot be used with cursor)
        page_cursor (str, optional): 페이징 처리를 위한 cursor (date가 없는 경우 필수이며 date와 함께 사용 불가)
                                     Cursor for pagination (required if date is not present and cannot be used with date)
    """

    if result_count < 10 or result_count > 1000:
        raise ValueError("result_count must be between 10 and 1000")

    validate_date_and_cursor(date, page_cursor)

    path = "/maplestory/v1/history/starforce"
    query = {
        "count": result_count,
        "date": kst.validate(date),
        "cursor": page_cursor,
    }
    response = fetch(path, query)
    return StarforceHistory.model_validate(response)



from pydantic import BaseModel

from maplestory.types.union import UnionWorldName


class UnionRankingInfo(BaseModel):
    """유니온 랭킹 상세 정보

    Attributes:
        date: 랭킹 업데이트 일자 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        ranking: 유니온 랭킹 순위
        character_name: 캐릭터 명
        world_name: 월드 명
        class_name: 직업 명
        sub_class_name: 전직 직업 명
        union_level: 유니온 레벨
        union_power: 유니온 파워
    """

    date: str
    ranking: int
    character_name: str
    world_name: UnionWorldName
    class_name: str
    sub_class_name: str
    union_level: int
    union_power: int


class UnionRanking(BaseModel):
    """유니온 랭킹 정보

    Attributes:
        ranking: 유니온 랭킹 정보
    """

    ranking: list[UnionRankingInfo]



from datetime import datetime

from pydantic import BaseModel, Field


class CharacterStatInfo(BaseModel):
    """캐릭터 스탯 상세 정보

    Attributes:
        name (str): 스탯 이름
        value (str): 스탯 값
    """

    name: str = Field(alias="stat_name")
    value: int | float = Field(alias="stat_value")


class CharacterStat(BaseModel):
    """캐릭터 스탯 정보

    Attributes:
        date (datetime): 조회 기준일
        character_class (str): 직업
        stats (list[CharacterStatInfo]): 현재 스탯 정보
        remain_ap (int): 잔여 AP
    """

    date: datetime = Field(repr=False)
    character_class: str
    stats: list[CharacterStatInfo] = Field(alias="final_stat")
    remain_ap: int



from datetime import datetime

from pydantic import BaseModel, Field


class HexaMatrixStatCore(BaseModel):
    """HEXA 코어 스탯 정보

    Attributes:
        slot_id (str): 슬롯 인덱스
        main_stat_name (str): 메인 스탯 명
        sub_stat_name_1 (str): 첫 번째 서브 명
        sub_stat_name_2 (str): 두 번째 서브 명
        main_stat_level (int): 메인 스탯 레벨
        sub_stat_level_1 (int): 첫 번째 서브 레벨
        sub_stat_level_2 (int): 두 번째 서브 레벨
        stat_grade (int): 스탯 코어 등급
    """

    slot_id: str
    main_stat_name: str
    sub_stat_name_1: str
    sub_stat_name_2: str
    main_stat_level: int
    sub_stat_level_1: int
    sub_stat_level_2: int
    stat_grade: int


class HexaMatrixStat(BaseModel):
    """HEXA 코어 스탯

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        character_class (str | None): 캐릭터 직업
        character_hexa_stat_core (list[HexaMatrixStatCore]): HEXA 코어 스탯 정보
        preset_hexa_stat_core (list[HexaMatrixStatCore]): HEXA 코어 스탯 정보 (프리셋)

    Notes:
        HEXA 코어 스탯 정보가 없는 경우 character_class가 None으로 표기됩니다.
    """

    date: datetime = Field(repr=False)
    character_class: str | None
    character_hexa_stat_core: list[HexaMatrixStatCore]
    preset_hexa_stat_core: list[HexaMatrixStatCore]



from .artifact import UnionArtifact, UnionArtifactCrystal, UnionArtifactEffect
from .raider import (
    UnionRaider,
    UnionRaiderBlock,
    UnionRaiderBlockControlPoint,
    UnionRaiderBlockPosition,
    UnionRaiderInnerStat,
)
from .union import UnionInfo

__all__ = [
    "UnionInfo",
    "UnionArtifact",
    "UnionArtifactCrystal",
    "UnionArtifactEffect",
    "UnionRaider",
    "UnionRaiderBlock",
    "UnionRaiderBlockControlPoint",
    "UnionRaiderBlockPosition",
    "UnionRaiderInnerStat",
]



from pydantic import BaseModel


class DojangRankingInfo(BaseModel):
    """무릉도장 랭킹 상세 정보

    Attributes:
        date: 랭킹 업데이트 일자 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        ranking: 무릉도장 랭킹 순위
        character_name: 캐릭터 명
        world_name: 월드 명
        class_name: 직업 명
        sub_class_name: 전직 직업 명
        character_level: 캐릭터 레벨
        dojang_floor: 무릉도장 층수
        dojang_time_record: 무릉도장 클리어 시간 기록 (초 단위)
    """

    date: str
    ranking: int
    character_name: str
    world_name: str
    class_name: str
    sub_class_name: str
    character_level: int
    dojang_floor: int
    dojang_time_record: int


class DojangRanking(BaseModel):
    """무릉도장 랭킹 정보

    Attributes:
        ranking: 무릉도장 랭킹 정보
    """

    ranking: list[DojangRankingInfo]



from pydantic import BaseModel


class AchievementRankingInfo(BaseModel):
    """업적 랭킹 상세 정보

    Attributes:
        date: 랭킹 업데이트 일자 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        ranking: 업적 랭킹 순위
        character_name: 캐릭터 명
        world_name: 월드 명
        class_name: 직업 명
        sub_class_name: 전직 직업 명
        trophy_grade: 업적 등급
        trophy_score: 업적 점수
    """

    date: str
    ranking: int
    character_name: str
    world_name: str
    class_name: str
    sub_class_name: str
    trophy_grade: str
    trophy_score: int


class AchievementRanking(BaseModel):
    """업적 랭킹 정보

    Attributes:
        ranking: 업적 랭킹 정보
    """

    ranking: list[AchievementRankingInfo]



"""이 모듈은 네트워크 요청을 처리하는 함수를 제공합니다. 

- `fetch` 함수는 주어진 URL 경로와 쿼리 파라미터를 사용하여 GET 요청을 보내고, 응답을 JSON 형태로 반환합니다. 
    만약 응답에 에러가 포함되어 있으면, `APIError`를 발생시킵니다.
"""

import httpx
from httpx import Response
from tenacity import retry, retry_if_exception_message, wait_exponential

from maplestory.config import Config
from maplestory.error import APIError, ErrorMessage


@retry(
    retry=retry_if_exception_message(match="^.*Please try again later.*$"),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
def fetch(url_path: str, query_params: dict | None = None) -> dict:
    """
    주어진 URL 경로와 쿼리 파라미터를 사용하여 GET 요청을 보내고, 응답을 JSON 형태로 반환합니다. 만약 응답에 에러가 포함되어 있으면, `APIError`를 발생시킵니다.

    Args:
        url_path (str): 요청을 보낼 URL 경로입니다.
        query_params (dict | None, optional): 요청에 포함할 쿼리 파라미터입니다. 기본값은 None입니다.

    Raises:
        APIError: 응답에 에러가 포함되어 있을 경우 발생합니다.

    Returns:
        dict: 응답을 JSON 형태로 변환한 결과입니다.
    """

    config = Config()
    resp: Response = httpx.get(
        f"{config.url}{url_path}",
        params=query_params,
        headers={
            "x-nxopen-api-key": config.key,
            "Accept": "application/json",
        },
        timeout=config.timeout,
    )

    data = resp.json()
    if error := data.get("error"):
        em = ErrorMessage.model_validate(error)
        raise APIError(em, resp)

    return data



"""유니온 정보 조회 API를 제공하는 모듈입니다.

Note:
    - 2023년 12월 21일 데이터부터 조회할 수 있습니다.
    - 유니온 정보 조회 API는 일자별 데이터로 매일 오전 1시부터 전일 데이터 조회가 가능합니다.
      (예를 들어, 12월 22일 데이터를 조회하면 22일 00시부터 23일의 00시 사이의 데이터가 조회됩니다.)
    - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
"""

from datetime import datetime

import maplestory.utils.date as dates
import maplestory.utils.kst as kst
from maplestory.models.union import UnionArtifact, UnionInfo, UnionRaider
from maplestory.utils.network import fetch


def get_union_info_by_ocid(
    character_ocid: str,
    date: datetime = kst.yesterday(),
) -> UnionInfo:
    """유니온 레벨 및 유니온 등급 정보를 조회합니다.

    Args:
        character_ocid: 캐릭터의 식별자(ocid).
        date: 조회 기준일 (KST).

    Returns:
        Union: 유니온의 레벨 및 등급 정보.
    """

    path = "/maplestory/v1/user/union"
    query = {
        "ocid": character_ocid,
        "date": dates.to_string(date),
    }
    response = fetch(path, query)

    return UnionInfo.model_validate(response)


def get_union_raider_info_by_ocid(
    character_ocid: str,
    date: datetime = kst.yesterday(),
) -> UnionRaider:
    """
    유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등 상세 정보를 조회합니다.

    Args:
        character_ocid: 캐릭터의 식별자(ocid).
        date: 조회 기준일 (KST).

    Returns:
        UnionRaider: 유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등의 상세 정보.
    """

    path = "/maplestory/v1/user/union-raider"
    query = {
        "ocid": character_ocid,
        "date": dates.to_string(date),
    }
    response = fetch(path, query)

    return UnionRaider.model_validate(response)


def get_union_artifact_info_by_ocid(
    character_ocid: str,
    date: datetime = kst.yesterday(),
) -> UnionArtifact:
    """
    유니온 아티팩트 정보를 조회합니다.

    Args:
        character_ocid: 캐릭터의 식별자(ocid).
        date: 조회 기준일 (KST).

    Returns:
        UnionArtifact: 유니온 아티팩트 정보.
    """

    path = "/maplestory/v1/user/union-artifact"
    query = {
        "ocid": character_ocid,
        "date": dates.to_string(date),
    }
    response = fetch(path, query)

    return UnionArtifact.model_validate(response)



from .achievement import AchievementRanking, AchievementRankingInfo
from .dojang import DojangRanking, DojangRankingInfo
from .guild import GuildRanking, GuildRankingInfo
from .overall import OverallRanking, OverallRankingInfo
from .theseed import TheSeedRanking, TheSeedRankingInfo
from .union import UnionRanking, UnionRankingInfo

__all__ = [
    "AchievementRanking",
    "AchievementRankingInfo",
    "DojangRanking",
    "DojangRankingInfo",
    "GuildRanking",
    "GuildRankingInfo",
    "OverallRanking",
    "OverallRankingInfo",
    "TheSeedRanking",
    "TheSeedRankingInfo",
    "UnionRanking",
    "UnionRankingInfo",
]



from __future__ import annotations

from datetime import datetime as _datetime
from datetime import timedelta
from typing import Annotated, Any
from zoneinfo import ZoneInfo

from annotated_types import Timezone
from pydantic import GetCoreSchemaHandler
from pydantic.types import _check_annotated_type
from pydantic_core import core_schema

KST_TZ = ZoneInfo("Asia/Seoul")
KSTdatetime = Annotated[_datetime, Timezone("Asia/Seoul")]


class AwareDatetime:
    """A datetime that requires KST timezone info."""

    KST_TZ_CONSTRAINT = 9 * 60 * 60

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        if cls is source:
            # used directly as a type
            return core_schema.datetime_schema(tz_constraint=cls.KST_TZ_CONSTRAINT)

        schema = handler(source)
        _check_annotated_type(schema["type"], "datetime", cls.__name__)
        schema["tz_constraint"] = cls.KST_TZ_CONSTRAINT
        return schema

    def __repr__(self) -> str:
        return "KstAwareDatetime"


def validate(date: _datetime):
    if date.tzinfo is None:
        raise ValueError("datetime should have timezone info.")

    if date.tzinfo.utcoffset(date) != timedelta(
        seconds=KST_TZ_CONSTRAINT
    ):
        raise ValueError("datetime should have KST timezone info.")


def now() -> AwareDatetime:
    """Returns the current datetime in the KST (Korea Standard Time) timezone.

    Returns:
        datetime: The current datetime in the KST timezone.
    """

    return _datetime.now(KST_TZ)


def yesterday() -> AwareDatetime:
    """
    한국 표준시(KST) 기준으로 어제의 날짜를 구합니다.
    Returns the datetime of yesterday in the KST (Korea Standard Time) timezone.

    Returns:
        datetime: 어제의 날짜와 시간(한국 표준시).
                  The datetime of yesterday in the KST timezone.

    Description:
        현재 한국 표준시(KST)를 기준으로 하루를 뺀 날짜를 반환합니다.
        시간대는 'Asia/Seoul'로 설정됩니다.
    """

    today_kst = _datetime.now(KST_TZ)
    return today_kst - timedelta(days=1)


def datetime(
    year, month=None, day=None, hour=0, minute=0, second=0, microsecond=0
) -> AwareDatetime:
    return _datetime(year, month, day, hour, minute, second, microsecond, tzinfo=KST_TZ)



from datetime import datetime

from PIL import Image
from pydantic import BaseModel, ConfigDict, Field, computed_field

from maplestory.services.guild import Guild
from maplestory.utils.image import get_image_from_url


class CharacterBasic(BaseModel):
    """캐릭터 기본 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        name (str): 캐릭터 명
        world (str): 월드 명
        gender (str): 캐릭터 성별
        class (str): 캐릭터 직업
        class_level (int): 캐릭터 직업 차수
        level (int): 캐릭터 레벨
        exp (int): 현재 레벨에서 보유한 경험치
        exp_rate (str): 현재 레벨에서 경험치 퍼센트
        guild_name (str | None): 캐릭터 소속 길드 명
        image (str): 캐릭터 외형 이미지
    """

    date: datetime = Field(repr=False)
    name: str = Field(alias="character_name")
    world: str = Field(alias="world_name")
    gender: str = Field(alias="character_gender")
    job: str = Field(alias="character_class")
    job_level: int = Field(alias="character_class_level")
    level: int = Field(alias="character_level")
    exp: int = Field(alias="character_exp")
    exp_rate: float = Field(alias="character_exp_rate")
    guild_name: str | None = Field(alias="character_guild_name")
    image_url: str = Field(alias="character_image", repr=False)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def exp_str(self) -> str:
        return f"{self.exp}({self.exp_rate})"

    @property
    def guild(self) -> Guild:
        return Guild(name=self.guild_name, world=self.world_name)

    @property
    def 조회기준일(self) -> datetime:
        return self.date

    @property
    def 캐릭터명(self) -> str:
        return self.character_name

    @property
    def 월드명(self) -> str:
        return self.world_name

    @property
    def 캐릭터성별(self) -> str:
        return self.character_gender

    @property
    def 캐릭터직업(self) -> str:
        return self.character_class

    @property
    def 캐릭터직업차수(self) -> int:
        return self.character_class_level

    @property
    def 레벨(self) -> int:
        return self.character_level

    @property
    def 현재레벨경험치(self) -> int:
        return self.character_exp

    @property
    def 현재레벨경험치퍼센트(self) -> str:
        return self.character_exp_rate

    @property
    def 길드(self) -> str:
        return self.character_guild_name

    @computed_field
    def image(self) -> Image.Image:
        return get_image_from_url(self.image_url)

    @property
    def 이미지(self) -> str:
        return self.character_image



from datetime import datetime

from pydantic import BaseModel, Field


class UnionInfo(BaseModel):
    """유니온 정보

    Attributes:
        date: 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        union_level: 유니온 레벨
        union_grade: 유니온 등급

    Examples:
        >>> Union(date="2024-01-04T00:00+09:00", union_level=9254, union_grade="그랜드 마스터 유니온 3")
        Union(date=datetime.datetime(2024, 1, 4, 0, 0, tzinfo=TzInfo(+09:00)),
              union_level=9254,
              union_grade='그랜드 마스터 유니온 3')
    """

    date: datetime = Field(repr=False)
    level: int = Field(alias="union_level")
    grade: str = Field(alias="union_grade")

    @property
    def date_string(self) -> str:
        return self.date.isoformat()



from pydantic_settings import BaseSettings, SettingsConfigDict


class Config(BaseSettings):
    key: str  # API Key
    url: str = "https://open.api.nexon.com"
    timeout: int = 5000

    model_config = SettingsConfigDict(
        env_file_encoding="utf-8",
        env_file=".env",
        env_prefix="MAPLESTORY_OPENAPI_",
        extra="ignore",
    )



from pydantic import BaseModel

# from pydantic.typing import ReprArgs


class HideNoneRepresentation:
    def __repr_args__(self: BaseModel):  # -> "ReprArgs":
        return [
            (key, value) for key, value in self.__dict__.items() if value is not None
        ]
        # for k, v in self.__dict__.items():
        #     if v is not None:
        #         yield k, v



"""확률 정보 조회 API를 제공하는 모듈입니다.

Note:
    - 큐브, 스타포스, 잠재능력 각각 조회 가능한 날짜가 다릅니다.
"""

from datetime import datetime

import maplestory.utils.date as dates
import maplestory.utils.kst as kst
from maplestory.enums import QueryableDate
from maplestory.models.history import Account, CubeHistory, StarforceHistory
from maplestory.models.history.potential import PotentialHistory
from maplestory.models.types import PageCursor
from maplestory.utils.network import fetch


def get_account_id() -> str:
    """계정 식별자(ouid) 조회

    Returns:
        str: 계정 식별자(ouid)
    """

    path = "/maplestory/v1/ouid"
    response = fetch(path)

    return Account.model_validate(response).ouid


def validate_date_and_cursor(
    date: datetime | None = None,
    cursor: PageCursor | None = None,
) -> None:
    """date와 cursor의 유효성을 확인하는 함수입니다.

    Args:
        date (datetime): 날짜 정보.
        cursor (str): 페이지 커서 정보.

    Raises:
        ValueError: date와 cursor가 동시에 제공되는 경우 또는 둘 다 제공되지 않는 경우 발생합니다.
    """

    if date is not None and cursor is not None:
        raise ValueError("date and cursor cannot be used together")

    if date is None and cursor is None:
        raise ValueError("date or cursor must be provided")


def validate_count(count: int) -> None:
    """count의 유효성을 확인하는 함수입니다.

    Args:
        count (int): 한번에 가져오려는 결과의 갯수.

    Raises:
        ValueError: count가 10보다 작거나 1000보다 클 경우 발생합니다.
    """

    if count < 10 or count > 1000:
        raise ValueError("count must be between 10 and 1000")


def get_starforce_history(
    result_count: int = 1000,
    date: datetime | None = kst.yesterday(),
    page_cursor: PageCursor | None = None,
) -> StarforceHistory:
    """스타포스 강화 결과를 조회합니다.

    Args:
        result_count (int): 한번에 가져오려는 결과의 갯수. (최소 10, 최대 1000)
        date (datetime, optional): 조회 기준일(KST). (cursor가 없는 경우 필수이며 cursor와 함께 사용 불가)
        page_cursor (str, optional): 페이징 처리를 위한 cursor. (date가 없는 경우 필수이며 date와 함께 사용 불가)

    Notes:
        - 스타포스 확률 정보는 최대 5분 후 확인 가능합니다.
        - 2023년 12월 27일 데이터부터 조회할 수 있습니다.
    """

    validate_count(result_count)

    if page_cursor is not None:
        date = None

    validate_date_and_cursor(date, page_cursor)

    if isinstance(date, datetime):
        dates.is_valid(date, QueryableDate.스타포스)

    path = "/maplestory/v1/history/starforce"
    query = {
        "count": result_count,
        "date": dates.to_string(date),
        "cursor": page_cursor,
    }
    response = fetch(path, query)

    return StarforceHistory.model_validate(response)


def get_cube_usage_history(
    result_count: int = 1000,
    date: datetime = kst.yesterday(),
    page_cursor: PageCursor | None = None,
) -> CubeHistory:
    """큐브 사용 결과를 조회합니다.

    Args:
        result_count (int): 한번에 가져오려는 결과의 갯수.
        date (datetime): 조회 기준일(KST).
        page_cursor (str) : 페이징 처리를 위한 cursor.

    Returns:
        CubeHistory: 큐브 사용 결과.

    Notes:
        - 데이터는 매일 오전 4시, 전일 데이터가 갱신됩니다.
        - 2022년 11월 25일 데이터부터 조회할 수 있습니다.
    """

    validate_count(result_count)

    if page_cursor is not None:
        date = None

    validate_date_and_cursor(date, page_cursor)

    if isinstance(date, datetime):
        dates.is_valid(date, QueryableDate.큐브)

    path = "/maplestory/v1/history/cube"
    query = {
        "count": result_count,
        "date": dates.to_string(date),
        "cursor": page_cursor,
    }
    response = fetch(path, query)

    return CubeHistory.model_validate(response)


def get_potential_history(
    result_count: int = 1000,
    date: datetime = kst.yesterday(),
    page_cursor: PageCursor | None = None,
) -> PotentialHistory:
    """잠재능력 재설정 이용 결과를 조회합니다.

    Attributes:
        count: 한번에 가져오려는 결과의 갯수(최소 10, 최대 1000)
        date: 조회 기준일 (KST) (cursor가 없는 경우 필수이며 cursor와 함께 사용 불가)
        cursor: 페이징 처리를 위한 cursor (date가 없는 경우 필수이며 date와 함께 사용 불가)

    Notes:
        - 데이터는 매일 오전 4시, 전일 데이터가 갱신됩니다.
        - 2024년 1월 25일 데이터부터 조회할 수 있습니다.
    """

    validate_count(result_count)

    if page_cursor is not None:
        date = None

    validate_date_and_cursor(date, page_cursor)

    if isinstance(date, datetime):
        dates.is_valid(date, QueryableDate.잠재능력)

    path = "/maplestory/v1/history/potential"
    query = {
        "count": result_count,
        "date": dates.to_string(date),
        "cursor": page_cursor,
    }
    response = fetch(path, query)

    return PotentialHistory.model_validate(response)



"""
Module This for interacting module provides with functionality for MapleStory guild interacting with MapleStory information.

This module provides functionality to fetch guilds.

and retrieve The guild information `Guild` class represents a from the MapleStory guild and provides various properties to access guild API.

Classes:
- Guild: Represents a information.
The guild in MapleStory.

`get_id` function Functions:
fetches the - guild identifier get_id: Fetches the (gcid) information guild identifier based on the guild name (gcid) information.
and world - name.
The get_basic_info: Fetches the `get_basic_info` basic function information fetches of the a basic information guild.

"""

from datetime import datetime

from pydantic import BaseModel, Field, computed_field

import maplestory.utils.date as dates
import maplestory.utils.kst as kst
from maplestory.apis.character import get_basic_character_info
from maplestory.models.character.basic import CharacterBasic
from maplestory.models.guild import GuildBasic, GuildModel
from maplestory.models.guild.basic import GuildSkill
from maplestory.utils.network import fetch


class Guild(BaseModel):
    name: str = Field(frozen=True)
    world: str = Field(frozen=True)

    @computed_field
    def id(self) -> str:
        return get_id(self.name, self.world).id

    def oguild_id(self) -> str:
        return self.id

    @computed_field
    def basic(self) -> GuildBasic:
        return get_basic_info(self.id)

    @property
    def level(self) -> int:
        return self.basic.level

    @property
    def point(self) -> int:
        return self.basic.point

    @property
    def fame(self) -> int:
        return self.basic.fame

    @property
    def member_count(self) -> int:
        return self.basic.member_count

    @property
    def members(self) -> list[str]:
        # TODO: Add character model
        return self.basic.members

    @property
    def master_name(self) -> str:
        return self.basic.master_name

    @property
    def master(self) -> str:
        # TODO: Add character model
        return self.basic.master_name

    @property
    def master_character(self) -> str:
        # TODO: Add character model
        return self.basic.master_character

    @property
    def skills(self) -> list[GuildSkill]:
        return self.basic.skills

    @property
    def noblesse_skills(self) -> list[GuildSkill]:
        return self.basic.noblesse_skills

    @property
    def mark(self) -> str:
        return self.basic.mark

    @property
    def custom_mark(self) -> str:
        return self.basic.custom_mark


def get_id(guild_name: str, world_name: str) -> str:
    """
    길드 식별자(gcid) 정보를 조회합니다.
    Fetches the guild identifier (gcid) information.

    Args:
        guild_name : 길드 명. The name of the guild.
        world_name : 월드 명. The name of the world.

    Returns:
        Guild: 길드 식별자(gcid) 정보. The guild identifier (gcid) information.
    """

    path = "/maplestory/v1/guild/id"
    query = {
        "guild_name": guild_name,
        "world_name": world_name,
    }
    response = fetch(path, query)

    return GuildModel.model_validate(response).id


def get_basic_info_by_id(
    guild_id: str,
    date: datetime = kst.yesterday(),
) -> GuildBasic:
    """
    길드 기본 정보를 조회합니다.
    Fetches the basic information of the guild.

    Args:
        guild_id : 길드 식별자. The identifier of the guild.
        date : 조회 기준일(KST). Reference date for the query (KST).

    Returns:
        GuildBasic: 길드의 기본 정보. The basic information of the guild.
    """

    path = "/maplestory/v1/guild/basic"
    query = {
        "oguild_id": guild_id,
        "date": dates.to_string(kst.validate(date)),
    }
    response = fetch(path, query)

    return GuildBasic.model_validate(response)


def get_basic_info(
    guild_name: str,
    world_name: str,
    date: datetime = kst.yesterday(),
) -> GuildBasic:
    """
    길드 기본 정보를 조회합니다.
    Fetches the basic information of the guild.

    Args:
        guild_name : 길드 명. The name of the guild.
        world_name : 월드 명. The name of the world.
        date : 조회 기준일(KST). Reference date for the query (KST).

    Returns:
        GuildBasic: 길드의 기본 정보. The basic information of the guild.
    """

    guild_id = get_id(guild_name, world_name)
    return get_basic_info_by_id(guild_id, date)


def get_basic_info_in_details(
    guild_name: str,
    world_name: str,
    date: datetime = kst.yesterday(),
) -> GuildBasic:
    """
    길드 기본 정보를 조회합니다.
    Fetches the basic information of the guild.

    Args:
        guild_name : 길드명. The name of the guild.
        world_name : 월드명. The name of the world.
        date : 조회 기준일(KST). Reference date for the query (KST).

    Returns:
        GuildBasic: 길드의 기본 정보. The basic information of the guild.
    """

    guild_basic: GuildBasic = get_basic_info(guild_name, world_name, date)

    # Add master character info
    guild_basic.master: CharacterBasic = get_basic_character_info(
        guild_basic.master_name, date
    )

    # TODO: Add members character info

    return guild_basic



import base64
from datetime import datetime
from io import BytesIO

from PIL import Image
from pydantic import BaseModel, Field, HttpUrl

from ..types import GuildLevel, WorldName


class GuildSkill(BaseModel):
    """길드 스킬 정보. Guild skill information.

    Attributes:
        name (str): 스킬명. The name of the guild skill.
        description (str): 스킬 설명. The description of the guild skill.
        level (int): 스킬 레벨. The level of the guild skill.
        effect (str): 스킬 레벨별 효과. The effect of the guild skill.
        icon (HttpUrl): 스킬 아이콘. The icon of the guild skill.
    """

    name: str = Field(alias="skill_name")
    description: str = Field(alias="skill_description")
    level: int = Field(alias="skill_level")
    effect: str = Field(alias="skill_effect")
    icon: HttpUrl = Field(alias="skill_icon")

    @property
    def desc(self) -> str:
        return self.description

    @property
    def 이름(self) -> str:
        return self.name

    @property
    def 설명(self) -> str:
        return self.description

    @property
    def 레벨(self) -> int:
        return self.level

    @property
    def 효과(self) -> str:
        return self.effect

    @property
    def 아이콘(self) -> HttpUrl:
        return self.icon


class GuildBasic(BaseModel):
    """길드 기본 정보. Basic information of the guild.

    Attributes:
        date: 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        world: 월드 명
        name: 길드 명
        level: 길드 레벨
        fame: 길드 명성치
        point: 길드 포인트(GP)
        master_name: 길드 마스터 캐릭터 명
        member_count: 길드원 수
        members: 길드원 목록
        skill: 길드 스킬 목록
        noblesse_skill: 노블레스 스킬 목록
        mark: 조합형 길드 마크
        custom_mark: 커스텀 길드 마크 (base64 인코딩 형식)
    """

    date: datetime = Field(repr=False)
    world: WorldName = Field(alias="world_name")
    name: str = Field(alias="guild_name")
    level: GuildLevel = Field(alias="guild_level")
    fame: int = Field(alias="guild_fame")
    point: int = Field(alias="guild_point")
    master_name: str = Field(alias="guild_master_name")
    member_count: int = Field(alias="guild_member_count")
    members: list[str] = Field(alias="guild_member")
    skills: list[GuildSkill] = Field(alias="guild_skill")
    noblesse_skills: list[GuildSkill] = Field(alias="guild_noblesse_skill")
    mark: str | None = Field(alias="guild_mark")
    custom_mark: str | None = Field(alias="guild_mark_custom")

    @property
    def custom_mark_img(self) -> Image.Image | None:
        if self.custom_mark is None:
            return None

        im_bytes = base64.b64decode(self.custom_mark)
        im_file = BytesIO(im_bytes)
        return Image.open(im_file)

    @property
    def 노블포인트(self) -> int:
        return sum(skill.level for skill in self.noblesse_skills)



from datetime import datetime

from pydantic import BaseModel, Field, field_validator


class CharacterSkillInfo(BaseModel):
    """스킬 정보

    Attributes:
        name (str): 스킬 명
        description (str): 스킬 설명
        level (int): 스킬 레벨
        effect (str | None): 스킬 레벨 별 효과 설명
        icon (str): 스킬 아이콘
    """

    name: str = Field(alias="skill_name")
    description: str = Field(alias="skill_description")
    level: int = Field(alias="skill_level")
    effect: str | None = Field(alias="skill_effect")
    icon: str = Field(alias="skill_icon")


class CharacterSkill(BaseModel):
    """캐릭터 스킬 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        character_class (str): 캐릭터 직업
        character_skill_grade (int): 스킬 전직 차수
        skills (list[CharacterSkillInfo]): 스킬 정보
    """

    date: datetime = Field(repr=False)
    character_class: str
    character_skill_grade: int
    skills: list[CharacterSkillInfo] = Field(alias="character_skill")

    @field_validator("character_skill_grade", mode="before")
    @classmethod
    def skill_grade_must_be_between_0_and_6(cls, v: str | int) -> int:
        """스킬 전직 차수 유효성 검사
        스킬 전직 차수가 0과 6 사이인지 확인

        Raises:
            ValueError: 스킬 전직 차수가 0 미만이거나 6 초과인 경우, 정수로 변환할 수 없는 경우
        """
        try:
            v = int(v)
        except ValueError as e:
            raise ValueError("skill_grade must be an integer.") from e

        if v < 0 or v > 6:
            raise ValueError("skill_grade must be in 0 ~ 6.")

        return v



from datetime import datetime

from pydantic import BaseModel, Field

from ..types import (
    CharacterClass,
    CharacterGender,
    ExpireDate,
    ItemGender,
    PresetNumber,
)


class CashItemColoringPrism(BaseModel):
    """캐시 장비 컬러링프리즘 정보

    Attributes:
        color_range: 컬러링프리즘 색상 범위
        hue: 컬러링프리즘 색상
        saturation: 컬러링프리즘 채도
        value: 컬러링프리즘 명도

    Examples:
        "color_range": "전체 색상 계열",
        "hue": 307,
        "saturation": -70,
        "value": 15
    """

    color_range: str
    hue: int
    saturation: int
    value: int


class CashitemOption(BaseModel):
    """캐시 장비 옵션

    Attributes:
        type: 옵션 종류
        value: 옵션 값
    """

    # TODO: None Check
    type: str = Field(alias="option_type")
    value: str = Field(alias="option_value")


class BaseCashItemPresetInfo(BaseModel):
    """캐시 장비 프리셋 정보

    Attributes:
        part: 캐시 장비 부위 명
        slot: 캐시 장비 슬롯 위치
        name: 캐시 장비 명
        icon: 캐시 장비 아이콘
        description: 캐시 장비 설명
        option: 캐시 장비 옵션
        date_expire: 아이템 유효기간 (KST)
        date_option_expire: 아이템 옵션 유효기간 (KST, 시간 단위 데이터로 분은 일괄 0으로 표기)
        label: 캐시 장비 라벨 정보
        coloring_prism: 캐시 장비 컬러링프리즘 정보
        gender: 아이템 장착 가능 성별
    """

    part: str = Field(alias="cash_item_equipment_part")
    slot: str = Field(alias="cash_item_equipment_slot")
    name: str = Field(alias="cash_item_name")
    icon: str = Field(alias="cash_item_icon")
    description: str | None = Field(alias="cash_item_description")
    option: list[CashitemOption] = Field(alias="cash_item_option")
    date_expire: ExpireDate
    date_option_expire: ExpireDate
    label: str | None = Field(alias="cash_item_label")
    coloring_prism: CashItemColoringPrism | None = Field(
        alias="cash_item_coloring_prism"
    )
    gender: ItemGender = Field(alias="item_gender")


class CashItemPresetInfo(BaseCashItemPresetInfo):
    """캐시 장비 프리셋 정보

    Attributes:
        part: 캐시 장비 부위 명
        slot: 캐시 장비 슬롯 위치
        name: 캐시 장비 명
        icon: 캐시 장비 아이콘
        description: 캐시 장비 설명
        option: 캐시 장비 옵션
        date_expire: 아이템 유효기간 (KST)
        date_option_expire: 아이템 옵션 유효기간 (KST, 시간 단위 데이터로 분은 일괄 0으로 표기)
        label: 캐시 장비 라벨 정보
        coloring_prism: 캐시 장비 컬러링프리즘 정보
        item_gender: 아이템 장착 가능 성별
        base_preset_item_disable_flag: 다른 프리셋에서 장비 추가 장착 없이 1번 프리셋의 장비 공유를 비활성화 했는지 여부
    """

    base_preset_item_disable_flag: str | None = None


class CashitemEquipment(BaseModel):
    """캐시 장비 정보

    Attributes:
        date: 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
            "2023-12-21T00:00+09:00"
        character_gender: 캐릭터 성별
        character_class: 캐릭터 직업
        preset_no: 적용 중인 캐시 장비 프리셋 번호
        items: 장착 중인 캐시 장비
        preset_1: 1번 프리셋 장착 캐시 장비 정보
        preset_2: 2번 프리셋 장착 캐시 장비 정보
        preset_3: 3번 프리셋 장착 캐시 장비 정보
        additional_items: 제로인 경우 베타, 엔젤릭버스터인 경우 드레스 업 모드에서 장착 중인 캐시 장비
        additional_preset_1: 제로인 경우 베타, 엔젤릭버스터인 경우 드레스 업 모드의 1번 프리셋 장착 캐시 장비 정보
        additional_preset_2: 제로인 경우 베타, 엔젤릭버스터인 경우 드레스 업 모드의 2번 프리셋 장착 캐시 장비 정보
        additional_preset_3: 제로인 경우 베타, 엔젤릭버스터인 경우 드레스 업 모드의 3번 프리셋 장착 캐시 장비 정보
    """

    date: datetime = Field(repr=False)
    character_gender: CharacterGender
    character_class: CharacterClass
    preset_no: PresetNumber
    items: list[BaseCashItemPresetInfo] = Field(alias="cash_item_equipment_base")
    preset_1: list[CashItemPresetInfo] = Field(alias="cash_item_equipment_preset_1")
    preset_2: list[CashItemPresetInfo] = Field(alias="cash_item_equipment_preset_2")
    preset_3: list[CashItemPresetInfo] = Field(alias="cash_item_equipment_preset_3")
    additional_items: list[BaseCashItemPresetInfo] = Field(
        alias="additional_cash_item_equipment_base"
    )
    additional_preset_1: list[CashItemPresetInfo] = Field(
        alias="additional_cash_item_equipment_preset_1"
    )
    additional_preset_2: list[CashItemPresetInfo] = Field(
        alias="additional_cash_item_equipment_preset_2"
    )
    additional_preset_3: list[CashItemPresetInfo] = Field(
        alias="additional_cash_item_equipment_preset_3"
    )



from datetime import datetime

from PIL import Image
from pydantic import BaseModel, ConfigDict, Field, computed_field

from maplestory.apis.guild import get_basic_info_by_id, get_guild_id
from maplestory.models.guild import GuildBasic
from maplestory.models.guild.basic import GuildSkill
from maplestory.utils.image import get_image_from_base64str, get_image_from_url
from maplestory.utils.kst import yesterday
from maplestory.utils.number import korean_format_number


class Guild(BaseModel):
    name: str = Field(frozen=True)
    world: str = Field(frozen=True)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @computed_field
    def id(self) -> str:
        return get_guild_id(self.name, self.world)

    @property
    def oguild_id(self) -> str:
        return self.id

    @computed_field(repr=False)
    def basic(self) -> GuildBasic:
        return get_basic_info(self.name, self.world)

    @computed_field
    def level(self) -> int:
        return self.basic.level

    @computed_field
    def point(self) -> int:
        return self.basic.point

    @computed_field
    def point_str(self) -> int:
        return korean_format_number(self.point)

    @computed_field
    def fame(self) -> int:
        return self.basic.fame

    @computed_field
    def fame_str(self) -> int:
        return korean_format_number(self.fame)

    @computed_field
    def member_count(self) -> int:
        return self.basic.member_count

    @computed_field
    def member_names(self) -> list[str]:
        return self.basic.members

    @computed_field
    def master_name(self) -> str:
        return self.basic.master_name

    @computed_field
    def skills(self) -> list[GuildSkill]:
        return self.basic.skills

    @computed_field
    def noblesse_skills(self) -> list[GuildSkill]:
        return self.basic.noblesse_skills

    @computed_field
    def mark(self) -> Image.Image | None:
        if self.basic.mark:
            return get_image_from_url(self.basic.mark)
        elif self.basic.custom_mark:
            return get_image_from_base64str(self.basic.custom_mark)
        return None

    @computed_field
    def is_custom_mark(self) -> bool:
        return self.basic.custom_mark is not None


def get_basic_info(
    guild_name: str,
    world_name: str,
    date: datetime = yesterday(),
) -> GuildBasic:
    """
    길드 기본 정보를 조회합니다.
    Fetches the basic information of the guild.

    Args:
        guild_name : 길드 명. The name of the guild.
        world_name : 월드 명. The name of the world.
        date : 조회 기준일(KST). Reference date for the query (KST).

    Returns:
        GuildBasic: 길드의 기본 정보. The basic information of the guild.
    """

    guild_id = get_guild_id(guild_name, world_name)
    return get_basic_info_by_id(guild_id, date)



from datetime import datetime

from pydantic import BaseModel, Field

from maplestory.utils.repr import HideZeroStatRepresentation

from ..types import CharacterClass


class SymbolItem(HideZeroStatRepresentation, BaseModel):
    """캐릭터 심볼 장비 상세 정보

    Attributes:
        name (str): 심볼 이름
        icon (str): 심볼 아이콘
        description (str): 심볼 설명
        force (int): 심볼 포스 증가 수치
        level (int): 심볼 레벨
        STR (int): 심볼로 증가한 힘(STR)
        DEX (int): 심볼로 증가한 민첩(DEX)
        INT (int): 심볼로 증가한 지력(INT)
        LUK (int): 심볼로 증가한 운(LUK)
        HP (int): 심볼로 증가한 체력(HP)
        growth_count (int): 현재 보유 성장치
        require_growth_count (int): 다음 레벨까지 필요 성장치
    """

    name: str = Field(alias="symbol_name")
    icon: str = Field(alias="symbol_icon")
    description: str = Field(alias="symbol_description")
    force: int = Field(alias="symbol_force")
    level: int = Field(alias="symbol_level")
    STR: int = Field(alias="symbol_str")
    DEX: int = Field(alias="symbol_dex")
    INT: int = Field(alias="symbol_int")
    LUK: int = Field(alias="symbol_luk")
    HP: int = Field(alias="symbol_hp")
    growth_count: int = Field(alias="symbol_growth_count")
    require_growth_count: int = Field(alias="symbol_require_growth_count")


class SymbolEquipment(BaseModel):
    """캐릭터 심볼 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        character_class (CharacterClass): 캐릭터 직업
        symbol (list[SymbolItem]): 심볼 정보
    """

    date: datetime = Field(repr=False)
    character_class: CharacterClass
    symbol: list[SymbolItem]



from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field

from maplestory.types.character import AbilityNumber

from ..types import Grade, PresetNumber


class AbilityInfoItem(BaseModel):
    """어빌리티 한 줄 정보

    Attributes:
        number (int): 어빌리티 번호 (몇 번째 줄인지)
        grade (str): 어빌리티 등급
        value (str): 어빌리티 옵션 및 수치
    """

    number: AbilityNumber = Field(alias="ability_no", repr=False)
    grade: Grade = Field(alias="ability_grade")
    value: str = Field(alias="ability_value")


class AbilityPreset(BaseModel):
    """어빌리티 프리셋 정보

    Attributes:
        grade (str): 어빌리티 프리셋의 어빌리티 등급
        info (list[AbilityInfoItem]): 어빌리티 프리셋 정보
    """

    grade: Grade = Field(alias="ability_preset_grade")
    info: list[AbilityInfoItem] = Field(alias="ability_info")

    def model_post_init(self, __context: Any) -> None:
        self.info = sorted(self.info, key=lambda x: x.number)


class Ability(BaseModel):
    """어빌리티 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        grade (Grade): 어빌리티 등급
        info (list[AbilityInfoItem]): 어빌리티 정보
        remain_fame (int): 보유 명성치
        preset_no (PresetNumber): 적용 중인 어빌리티 프리셋 번호(number)
        preset1 (AbilityPreset): 어빌리티 1번 프리셋 정보
        preset2 (AbilityPreset): 어빌리티 2번 프리셋 정보
        preset3 (AbilityPreset): 어빌리티 3번 프리셋 정보
    """

    date: datetime = Field(repr=False)
    grade: Grade = Field(alias="ability_grade")
    info: list[AbilityInfoItem] = Field(alias="ability_info")
    remain_fame: int
    preset_no: PresetNumber
    preset1: AbilityPreset = Field(alias="ability_preset_1", repr=False)
    preset2: AbilityPreset = Field(alias="ability_preset_2", repr=False)
    preset3: AbilityPreset = Field(alias="ability_preset_3", repr=False)



from enum import Enum, Flag, auto

from kst.datetime import timedelta

import maplestory.utils.kst as kst


class BaseEnum(Enum):
    @classmethod
    def members(cls):
        return list(cls.__members__)

    @classmethod
    def keys(cls):
        return list(cls.__members__)

    @classmethod
    def values(cls):
        return [c.value for c in cls]

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class WorldType(BaseEnum):
    일반 = 0  # normal
    리부트 = 1  # reboot
    Normal = 0
    Reboot = 1


class GuildRankType(BaseEnum):
    주간명성치 = 0  # Weekly Fame
    플레그레이스 = 1  # Flag Race
    지하수로 = 2  # Underground Waterway


class MinimumDateType(BaseEnum):
    랭킹 = kst.datetime(2023, 12, 22)
    유니온 = 1  # Union
    길드 = 2  # Guild


class Planet(Enum):
    MERCURY = (3.303e23, 2.4397e6)
    VENUS = (4.869e24, 6.0518e6)
    EARTH = (5.976e24, 6.37814e6)
    MARS = (6.421e23, 3.3972e6)
    JUPITER = (1.9e27, 7.1492e7)
    SATURN = (5.688e26, 6.0268e7)
    URANUS = (8.686e25, 2.5559e7)
    NEPTUNE = (1.024e26, 2.4746e7)

    def __init__(self, mass, radius):
        self.mass = mass  # in kilograms
        self.radius = radius  # in meters

    @property
    def surface_gravity(self):
        # universal gravitational constant  (m3 kg-1 s-2)
        G = 6.67300e-11
        return G * self.mass / (self.radius * self.radius)


# >>> Planet.EARTH.value
# (5.976e+24, 6378140.0)
# >>> Planet.EARTH.surface_gravity
# 9.802652743337129


class Period(timedelta, Enum):
    "different lengths of time"
    _ignore_ = "Period i"
    Period = vars()
    for i in range(367):
        Period["day_%d" % i] = i


# >>> list(Period)[:2]
# [<Period.day_0: kst.datetime.timedelta(0)>, <Period.day_1: kst.datetime.timedelta(days=1)>]
# >>> list(Period)[-2:]
# [<Period.day_365: kst.datetime.timedelta(days=365)>, <Period.day_366: kst.datetime.timedelta(days=366)>]


class Color(Flag):
    RED = auto()
    GREEN = auto()
    BLUE = auto()


# >>> purple = Color.RED | Color.BLUE
# >>> white = Color.RED | Color.GREEN | Color.BLUE
# >>> Color.GREEN in purple
# False
# >>> Color.GREEN in white
# True
# >>> purple in white
# True
# >>> white in purple
# False


if __name__ == "__main__":
    from devtools import debug

    val = 1
    w = WorldType(val)
    debug(w)
    w = WorldType.일반
    debug(w)
    w = WorldType.Normal
    debug(w)
    query = {"world_type": w, "world_name": "테스트"}
    debug(query)

    debug(WorldType._member_names_)
    debug(WorldType._member_map_)
    debug(WorldType._value2member_map_)
    debug(WorldType.list())
    debug(WorldType.values())
    debug(WorldType.members())
    # debug(WorldType.__members__)
    # debug(list(WorldType.__members__))



"""
이 모듈은 메이플스토리 게임의 유니온 정보를 가져오고 처리하는 함수를 제공합니다.
This module provides functions to fetch and handle Union information in the MapleStory game.

이 모듈에는 주어진 캐릭터 식별자(ocid)와 참조 날짜에 대한 유니온 레벨과 유니온 등급 정보를 가져오는 `get_union_info` 함수가 포함되어 있습니다.
It includes a function `get_union_info` which fetches the Union level and Union rank information 
for a given character identifier (ocid) and reference date.

이 모듈은 `maplestory.models.union` 패키지의 `Union`과 `UnionRaider` 모델과 `maplestory.utils.date`와 `maplestory.utils.network` 패키지의 유틸리티 함수를 사용합니다.
The module uses the `Union` and `UnionRaider` models from the `maplestory.models.union` package 
and utility functions from the `maplestory.utils.date` and `maplestory.utils.network` packages.
"""
from __future__ import annotations

import re
from datetime import datetime
from typing import Any, Generator

from pydantic import BaseModel, Field, RootModel, computed_field, field_validator
from rich import print

from maplestory.apis.union import (
    get_union_artifact_info_by_ocid,
    get_union_info_by_ocid,
    get_union_raider_info_by_ocid,
)
from maplestory.models.union import UnionArtifact, UnionInfo, UnionRaider
from maplestory.models.union.artifact import UnionArtifactEffect
from maplestory.services.character import get_character_id
from maplestory.utils.kst import yesterday


class Union(BaseModel):
    character_name: str
    date: datetime = yesterday()

    # def model_post_init(self, __context: Any) -> None:
    #     union_info = get_union_info(self.character_name, self.date)
    #     # for key in union_info.model_fields_set:
    #     #     if key != 'date':
    #     #         setattr(self, key, getattr(union_info, key))
    #     self.level = union_info.level
    #     self.grade = union_info.grade

    #     union_raider = get_union_raider_info(self.character_name, self.date)
    #     self.raider = union_raider

    #     union_artifact = get_union_artifact_info(self.character_name, self.date)
    #     self.artifact = union_artifact

    # effects: list[UnionArtifactEffect] = Field(alias="union_artifact_effect")
    # crystals: list[UnionArtifactCrystal] = Field(alias="union_artifact_crystal")
    # remain_ap: int | None = Field(alias="union_artifact_remain_ap")

    @computed_field(repr=False)
    @property
    def union_info(self) -> UnionInfo:
        return get_union_info(self.character_name, self.date)

    @computed_field
    @property
    def level(self) -> int:
        return self.union_info.level

    @computed_field
    @property
    def grade(self) -> str:
        return self.union_info.grade

    @computed_field(repr=False)
    @property
    def raider_info(self) -> UnionRaider:
        return get_union_raider_info(self.character_name, self.date)

    @computed_field
    @property
    def raider_stats(self) -> UnionStats:
        stats = UnionStats.parse_list(self.raider_info.union_raider_stat)
        return stats.summarize(split_multiple=True)

    @computed_field
    @property
    def 공격대원효과(self) -> UnionStats:
        return self.union_raider_stat

    @computed_field
    @property
    def occupied_stats(self) -> UnionStats:
        stats = UnionStats.parse_list(self.raider_info.union_occupied_stat)
        return stats.summarize(split_multiple=True)

    @computed_field
    @property
    def 공격대점령효과(self) -> UnionStats:
        return self.union_occupied_stat

    @computed_field(repr=False)
    @property
    def artifact(self) -> UnionArtifact:
        return get_union_artifact_info(self.character_name, self.date)

    @computed_field
    @property
    def artifact_effects(self) -> list[UnionArtifactEffect]:
        return self.artifact.effects

    @computed_field
    @property
    def 아티팩트효과(self) -> list[UnionArtifactEffect]:
        return self.artifact_effects


def get_union_info(
    character_name: str,
    date: datetime = yesterday(),
) -> UnionInfo:
    """
    유니온 레벨 및 유니온 등급 정보를 조회합니다.
    Fetches the Union level and Union rank information.

    Args:
        character_name : 캐릭터의 이름.
                         The name of the character.
        date: 조회 기준일 (KST).
                        Reference date for the query (KST).

    Returns:
        Union: 유니온의 레벨 및 등급 정보.
               The level and rank information of the Union.
    """

    character_ocid = get_character_id(character_name)
    return get_union_info_by_ocid(character_ocid, date)


def get_union_raider_info(
    character_name: str,
    date: datetime = yesterday(),
) -> UnionRaider:
    """
    유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등 상세 정보를 조회합니다.
    Fetches the detailed information of the effect of the attack members placed in the Union and the effect of the attack occupation.

    Args:
        character_name : 캐릭터의 이름.
                         The name of the character.
        date: 조회 기준일 (KST).
                        Reference date for the query (KST).

    Returns:
        UnionRaider: 유니온에 배치된 공격대원 효과 및 공격대 점령 효과 등의 상세 정보.
                     The detailed information of the effect of the attack members placed in the Union and the effect of the attack occupation.
    """

    character_ocid = get_character_id(character_name)
    return get_union_raider_info_by_ocid(character_ocid, date)


# def summarize_union_stat(stat: str) -> dict[str, str]:
#     if m := re.search(r"(?P<name>.+) (?P<value>\d+)(?P<percent>%?) (?P<updown>.+)", stat):
#         return m.groupdict()


#     raise ValueError("Invalid stat format.")
def summarize_union_stat(stat: str) -> dict[str, str]:
    if m := re.search(
        r"(?P<name>.+) (?P<value>\d+(\.\d+)?)(?P<percent>%?) (?P<updown>.+)", stat
    ):
        return m.groupdict()

    raise ValueError("Invalid stat format.")


class UnionStat(BaseModel):
    stat: str = Field(repr=False)

    def __repr__(self) -> str:
        return f"UnionStat({self.stat!r})"

    def __rich_repr__(self) -> Generator[str]:
        yield self.stat

    @field_validator("stat", mode="before")
    @classmethod
    def unify_format(cls, stat: str) -> str:
        # '공격력 20, 마력 20 증가' -> '공격력/마력 20 증가'
        pattern = r"(\w+) (\d+), (\w+) (\d+) (\w+)"
        if m := re.match(pattern, stat):
            if m[2] == m[4]:
                replacement = r"\1/\3 \2 \5"
                return re.sub(pattern, replacement, stat)

        # '몬스터 방어율 무시 20% 증가' -> '방어율 무시 20% 증가'
        if "몬스터 방어율 무시" in stat:
            stat = stat.replace("몬스터 ", "")

        return stat

    @computed_field(repr=False)
    @property
    def parsed_stat(self) -> dict[str, str]:
        return summarize_union_stat(self.stat)

    @computed_field
    @property
    def name(self) -> str:
        return self.parsed_stat.get("name")

    @property
    def is_multi_name(self) -> bool:
        return "," in self.name or "/" in self.name

    @property
    def multi_name(self) -> str | None:
        if "," in self.name:
            # 'STR, DEX, LUK 40 증가'
            return [s.strip() for s in self.name.split(",")]
        elif "/" in self.name:
            # '공격력/마력 20 증가'
            return [s.strip() for s in self.name.split("/")]
        return None

    @computed_field
    @property
    def percent(self) -> str:
        return self.parsed_stat.get("percent")

    @computed_field
    @property
    def is_percent(self) -> bool:
        return self.percent == "%"

    @computed_field
    @property
    def value(self) -> int | float:
        try:
            return int(self.parsed_stat.get("value"))
        except ValueError as _:
            return float(self.parsed_stat.get("value"))

    @computed_field
    @property
    def updown(self) -> str:
        return self.parsed_stat.get("updown")

    def __add__(self, other: UnionStat) -> UnionStat:
        if self.name != other.name:
            raise ValueError("Cannot add different stats.")
        if self.updown != other.updown:
            raise ValueError("Cannot add stats with different updown.")
        if self.is_percent != other.is_percent:
            raise ValueError(
                "Cannot add stats with different types (percent vs number)."
            )

        new_value = self.value + other.value
        if self.is_percent:
            new_value = f"{new_value}%"
        return UnionStat(stat=f"{self.name} {new_value} {self.updown}")

    def __radd__(self, other: UnionStat) -> UnionStat:
        if other == 0:
            return self
        return self.__add__(other) if isinstance(other, UnionStat) else NotImplemented


class UnionStats(RootModel):
    root: list[UnionStat]

    def __repr__(self) -> str:
        return f"UnionStats({self.root!r})"

    def __rich_repr__(self) -> Generator:
        yield self.root

    def model_post_init(self, __context: Any) -> None:
        """
        모델 생성 후 초기화를 수행합니다. 스탯의 이름을 기준으로 root를 정렬합니다.

        Args:
            __context (Any): 초기화 후의 컨텍스트.
        """
        """
        Initializes the model post creation. Sorts the root based on the name of the stat.

        Args:
            __context (Any): The context for post initialization.
        """

        self.root = sorted(self.root, key=lambda x: x.name)

    def group_stats(self, split_multiple: bool = False) -> dict[str, list[UnionStat]]:
        """
        스탯의 이름과 퍼센트 여부를 기준으로 스탯을 그룹화합니다.
        split_multiple이 True인 경우, 다중 이름 스탯을 개별 스탯으로 분리합니다.

        Args:
            split_multiple (bool): 다중 이름 스탯을 분리할지 여부. 기본값은 False입니다.

        Returns:
            Dict[str, List[UnionStat]]: 그룹화된 스탯.
        """
        """
        Groups the stats based on the name and whether it is a percentage. 
        If split_multiple is True, splits the multi-name stats into individual stats.

        Args:
            split_multiple (bool): Whether to split multi-name stats. Default is False.

        Returns:
            Dict[str, List[UnionStat]]: The grouped stats.
        """
        stats = []
        if split_multiple:
            for stat in self.root:
                if stat.is_multi_name:
                    stats.extend(
                        [
                            UnionStat(
                                stat=f"{name} {stat.value}{stat.percent} {stat.updown}"
                            )
                            for name in stat.multi_name
                        ]
                    )
                else:
                    stats.append(stat)
        else:
            stats = self.root

        return {
            (stat.name, stat.is_percent): [
                s
                for s in stats
                if s.name == stat.name and s.is_percent == stat.is_percent
            ]
            for stat in stats
        }

    def group_with_split_multiple(self) -> dict[str, list[UnionStat]]:
        """
        split_multiple을 True로 설정하여 스탯을 그룹화합니다.

        Returns:
            Dict[str, List[UnionStat]]: 그룹화된 스탯.
        """
        """
        Groups the stats with split_multiple set to True.

        Returns:
            Dict[str, List[UnionStat]]: The grouped stats.
        """
        return self.group_stats(split_multiple=True)

    @classmethod
    def parse_list(cls, data: list[str]) -> UnionStats:
        """
        스탯의 리스트를 파싱하여 UnionStats 객체를 생성합니다.

        Args:
            data (List[str]): 스탯의 리스트.

        Returns:
            UnionStats: 생성된 UnionStats 객체.
        """
        """
        Parses a list of stats and creates a UnionStats object.

        Args:
            data (List[str]): The list of stats.

        Returns:
            UnionStats: The created UnionStats object.
        """
        return cls(root=[UnionStat(stat=stat) for stat in data])

    def summarize(self, split_multiple: bool = False) -> UnionStats:
        """
        스탯을 요약합니다. '방어율 무시'는 합이 아닌 곱으로 처리되므로 별도로 처리합니다.

        Args:
            split_multiple (bool): 다중 이름 스탯을 분리할지 여부. 기본값은 False입니다.

        Returns:
            UnionStats: 요약된 스탯.
        """
        """
        Summarizes the stats. '방어율 무시' (Ignore Defense Rate) is handled separately 
        as it is not summed but multiplied.

        Args:
            split_multiple (bool): Whether to split multi-name stats. Default is False.

        Returns:
            UnionStats: The summarized stats.
        """
        combined_stats = []
        grouped_stats = self.group_stats(split_multiple)

        for (stat_name, _), stats in grouped_stats.items():
            if stat_name == "방어율 무시":
                # '방어율 무시' (Ignore Defense Rate) is multiplied, not summed
                # 방어율 무시는 합적용이 아니라 곱적용이라 별도로 처리
                combined_stats.extend(stats)
            else:
                combined_stats.append(sum(stats))

        return UnionStats(root=combined_stats)


# (.+) (\d+%?) (증가|감소|회복)
# (.+) (\d+%?) (.+)

# 경험치 획득량 10% 증가
# 공격 시 20%의 확률로 데미지 20% 증가
# 공격력/마력 20 증가
# 메소 획득량 4% 증가
# 방어율 무시 6% 증가
# 버프 지속시간 25% 증가
# 보스 몬스터 공격 시 데미지 6% 증가
# 상태 이상 내성 4 증가
# 소환수 지속시간 10% 증가
# 스킬 재사용 대기시간 6% 감소
# 적 공격마다 70%의 확률로 순수 HP의 8% 회복
# 최대 HP 2000 증가
# 최대 HP 5% 증가
# 크리티컬 데미지 6% 증가
# 크리티컬 확률 4% 증가
# DEX 80 증가
# INT 80 증가
# LUK 80 증가
# STR 100 증가
# STR 80 증가
# STR, DEX, LUK 40 증가


def get_union_artifact_info(
    character_name: str,
    date: datetime = yesterday(),
) -> UnionArtifact:
    """
    유니온 아티팩트 정보를 조회합니다.
    Fetches the Union Artifact information.

    Args:
        character_name : 캐릭터의 이름.
                         The name of the character.
        date: 조회 기준일 (KST).
                        Reference date for the query (KST).

    Returns:
        UnionArtifact: 유니온 아티팩트 정보.
                       The information of the Union Artifact.
    """

    character_ocid = get_character_id(character_name)
    return get_union_artifact_info_by_ocid(character_ocid, date)


# result = summarize_union_stat("공격 시 20%의 확률로 데미지 20% 증가")
# print(result)


union_raider_stat = [
    "INT 80 증가",
    "보스 몬스터 공격 시 데미지 6% 증가",
    "최대 HP 5% 증가",
    "INT 80 증가",
    "소환수 지속시간 10% 증가",
    "STR 80 증가",
    "INT 80 증가",
    "크리티컬 확률 4% 증가",
    "적 공격마다 70%의 확률로 순수 HP의 8% 회복",
    "공격 시 20%의 확률로 데미지 20% 증가",
    "STR 80 증가",
    "크리티컬 데미지 6% 증가",
    "DEX 80 증가",
    "경험치 획득량 10% 증가",
    "DEX 80 증가",
    "최대 HP 2000 증가",
    "STR 80 증가",
    "상태 이상 내성 4 증가",
    "INT 80 증가",
    "STR 80 증가",
    "STR, DEX, LUK 40 증가",
    "최대 HP 2000 증가",
    "INT 80 증가",
    "STR 100 증가",
    "STR 100 증가",
    "STR 100 증가",
    "LUK 80 증가",
    "공격력/마력 20 증가",
    "LUK 80 증가",
    "DEX 80 증가",
    "크리티컬 확률 4% 증가",
    "STR 80 증가",
    "버프 지속시간 25% 증가",
    "방어율 무시 6% 증가",
    "DEX 80 증가",
    "메소 획득량 4% 증가",
    "스킬 재사용 대기시간 6% 감소",
    "DEX 80 증가",
    "LUK 80 증가",
    "STR 70 증가",
    "크리티컬 데미지 20.00% 증가",
    "크리티컬 확률 20% 증가",
    "방어율 무시 40% 증가",
    "최대 MP 250 증가",
    "공격력 9 증가",
    "보스 몬스터 공격 시 데미지 40% 증가",
    "마력 1 증가",
    "STR 70 증가",
    "일반 몬스터 공격 시 데미지 40% 증가",
    "마력 4 증가",
    "획득 경험치 10.0% 증가",
    "크리티컬 확률 40% 증가",
    "공격력 15 증가",
    "DEX 80 증가",
    "적 공격마다 70%의 확률로 순수 MP의 8% 회복",
    "LUK 80 증가",
    "적 공격마다 70%의 확률로 순수 HP의 8% 회복",
    "최대 MP 5% 증가",
    "최대 HP 2000 증가",
    "보스 몬스터 공격 시 데미지 5% 증가",
    "크리티컬 확률 4% 증가",
    "스킬 재사용 대기시간 5% 감소",
    "메소 획득량 4% 증가",
    "STR, DEX, LUK 40 증가",
    "LUK 80 증가",
    "STR 100 증가",
    "INT 80 증가",
    "공격력/마력 20 증가",
    "상태 이상 내성 4 증가",
    "크리티컬 확률 4% 증가",
    "LUK 80 증가",
    "DEX 80 증가",
    "INT 80 증가",
    "소환수 지속시간 10% 증가",
    "LUK 80 증가",
    "STR 80 증가",
    "INT 80 증가",
    "STR 80 증가",
    "버프 지속시간 20% 증가",
    "공격 시 20%의 확률로 데미지 16% 증가",
    "STR 80 증가",
    "STR 80 증가",
    "DEX 80 증가",
    "방어율 무시 5% 증가",
    "크리티컬 데미지 5% 증가",
    "LUK 80 증가",
    "INT 80 증가",
    "STR 80 증가",
    "STR 80 증가",
    "INT 80 증가",
    "크리티컬 데미지 20.00% 증가",
    "마력 5 증가",
    "LUK 5 증가",
    "크리티컬 확률 11% 증가",
    "보스 몬스터 공격 시 데미지 23% 증가",
    "INT 25 증가",
    "버프 지속시간 40% 증가",
    "방어율 무시 33% 증가",
    "크리티컬 데미지 2.40% 증가",
    "보스 몬스터 공격 시 데미지 15.00% 증가",
    "데미지 15.00% 증가",
    "공격력 18, 마력 18 증가",
    "크리티컬 데미지 6.50% 증가",
    "상태 이상 내성 13 증가",
]


# union_raider_stat = ["공격력 20, 마력 20 증가", "공격력/마력 18 증가"]

# us = UnionStats.parse_list(union_raider_stat)
# print(us)
# print(us.group_with_split_multiple())
# stats = us.root
# for stat in stats:
#     if stat.is_multi_name:
#         stats.remove(stat)
#         for name in stat.multi_name:
#             stats.append(UnionStat(stat=f"{name} {stat.value}{stat.percent} {stat.updown}"))
# print('-' * 10)
# print(stats)

# s = UnionStat(stat='공격력 20, 마력 20 증가')
# print(s)
# print(s.multi_name)
# print(s.is_multi_name)
# s = UnionStat(stat="공격력/마력 20 증가")
# print(s)
# print(s.multi_name)
# print(s.is_multi_name)

# print(s.is_multi_name)
# print(s.multi_name)

# for name in s.multi_name:
#     print(UnionStat(stat=f"{name} {stat.value}{stat.percent} {stat.updown}"))

# print('-' * 10)
# print(us.group_with_split_multiple())
# print(us.summarize(split_multiple=True))
# print(us.summarize())
# UnionStat('크리티컬 확률 8% 증가')

# lis = [
#     UnionStat(stat="DEX 80 증가"),
#     UnionStat(stat="DEX 80 증가"),
#     UnionStat(stat="DEX 80 증가"),
#     UnionStat(stat="DEX 80 증가"),
# ]
# print(sum(lis))
# # TypeError: unsupported operand type(s) for +: 'int' and 'UnionStat'

# result = UnionStat(stat="DEX 80 증가") + UnionStat(stat="DEX 80 증가")
# print(result)



from typing import Literal

ClassName = Literal[
    "초보자-전체 전직",
    "전사-전체 전직",
    "전사-검사",
    "전사-파이터",
    "전사-페이지",
    "전사-스피어맨",
    "전사-크루세이더",
    "전사-나이트",
    "전사-버서커",
    "전사-히어로",
    "전사-팔라딘",
    "전사-다크나이트",
    "마법사-전체 전직",
    "마법사-매지션",
    "마법사-위자드(불,독)",
    "마법사-위자드(썬,콜)",
    "마법사-클레릭",
    "마법사-메이지(불,독)",
    "마법사-메이지(썬,콜)",
    "마법사-프리스트",
    "마법사-아크메이지(불,독)",
    "마법사-아크메이지(썬,콜)",
    "마법사-비숍",
    "궁수-전체 전직",
    "궁수-아처",
    "궁수-헌터",
    "궁수-사수",
    "궁수-레인저",
    "궁수-저격수",
    "궁수-보우마스터",
    "궁수-신궁",
    "궁수-아처(패스파인더)",
    "궁수-에인션트아처",
    "궁수-체이서",
    "궁수-패스파인더",
    "도적-전체 전직",
    "도적-로그",
    "도적-어쌔신",
    "도적-시프",
    "도적-허밋",
    "도적-시프마스터",
    "도적-나이트로드",
    "도적-섀도어",
    "도적-세미듀어러",
    "도적-듀어러",
    "도적-듀얼마스터",
    "도적-슬래셔",
    "도적-듀얼블레이더",
    "해적-전체 전직",
    "해적-해적",
    "해적-인파이터",
    "해적-건슬링거",
    "해적-캐논슈터",
    "해적-버커니어",
    "해적-발키리",
    "해적-캐논블래스터",
    "해적-바이퍼",
    "해적-캡틴",
    "해적-캐논마스터",
    "기사단-전체 전직",
    "기사단-노블레스",
    "기사단-소울마스터",
    "기사단-플레임위자드",
    "기사단-윈드브레이커",
    "기사단-나이트워커",
    "기사단-스트라이커",
    "기사단-미하일",
    "아란-전체 전직",
    "에반-전체 전직",
    "레지스탕스-전체 전직",
    "레지스탕스-시티즌",
    "레지스탕스-배틀메이지",
    "레지스탕스-와일드헌터",
    "레지스탕스-메카닉",
    "레지스탕스-데몬슬레이어",
    "레지스탕스-데몬어벤져",
    "레지스탕스-제논",
    "레지스탕스-블래스터",
    "메르세데스-전체 전직",
    "팬텀-전체 전직",
    "루미너스-전체 전직",
    "카이저-전체 전직",
    "엔젤릭버스터-전체 전직",
    "초월자-전체 전직",
    "초월자-제로",
    "은월-전체 전직",
    "프렌즈 월드-전체 전직",
    "프렌즈 월드-키네시스",
    "카데나-전체 전직",
    "일리움-전체 전직",
    "아크-전체 전직",
    "호영-전체 전직",
    "아델-전체 전직",
    "카인-전체 전직",
    "라라-전체 전직",
    "칼리-전체 전직",
]



from datetime import datetime

from pydantic import BaseModel, Field


class HyperStatItem(BaseModel):
    """캐릭터 하이퍼스탯 한 줄 정보

    Attributes:
        type (str): 스탯 종류
        point (int | None): 스탯 투자 포인트
        level (int): 스탯 레벨
        increase (str | None): 스탯 상승량
    """

    type: str = Field(alias="stat_type")
    point: int | None = Field(alias="stat_point")
    level: int = Field(alias="stat_level")
    increase: str | None = Field(alias="stat_increase")


class HyperStat(BaseModel):
    """캐릭터 하이퍼스탯 정보

    Attributes:
        date (datetime): 조회 기준일
        character_class (str): 캐릭터 직업
        preset_no (str): 적용 중인 프리셋 번호
        use_available (int): 사용 가능한 최대 하이퍼스탯 포인트
        preset1 (list[HyperStatItem]): 프리셋 1번 하이퍼 스탯 정보
        preset1_remain_point (int): 프리셋 1번 하이퍼 스탯 잔여 포인트
        preset2 (list[HyperStatItem]): 프리셋 2번 하이퍼 스탯 정보
        preset2_remain_point (int): 프리셋 2번 하이퍼 스탯 잔여 포인트
        preset3 (list[HyperStatItem]): 프리셋 3번 하이퍼 스탯 정보
        preset3_remain_point (int): 프리셋 3번 하이퍼 스탯 잔여 포인트
    """

    date: datetime = Field(repr=False)
    character_class: str
    preset_no: str = Field(alias="use_preset_no")
    use_available: int = Field(alias="use_available_hyper_stat")
    preset1: list[HyperStatItem] = Field(alias="hyper_stat_preset_1")
    preset1_remain_point: int = Field(alias="hyper_stat_preset_1_remain_point")
    preset2: list[HyperStatItem] = Field(alias="hyper_stat_preset_2")
    preset2_remain_point: int = Field(alias="hyper_stat_preset_2_remain_point")
    preset3: list[HyperStatItem] = Field(alias="hyper_stat_preset_3")
    preset3_remain_point: int = Field(alias="hyper_stat_preset_3_remain_point")



from datetime import datetime

from pydantic import BaseModel, Field


class HairInfo(BaseModel):
    """캐릭터 헤어 정보

    Attributes:
        hair_name (str): 헤어 명
        base_color (str): 헤어 베이스 컬러
        mix_color (str | None): 헤어 믹스 컬러
        mix_rate (str): 헤어 믹스 컬러의 염색 비율
    """

    hair_name: str
    base_color: str
    mix_color: str | None
    mix_rate: str


class FaceInfo(BaseModel):
    """캐릭터 성형 정보

    Attributes:
        face_name (str): 성형 명
        base_color (str): 성형 베이스 컬러
        mix_color (str | None): 성형 믹스 컬러
        mix_rate (str): 성형 믹스 컬러의 염색 비율
    """

    face_name: str
    base_color: str
    mix_color: str | None
    mix_rate: str


class BeautyEquipment(BaseModel):
    """캐릭터 헤어, 성형, 피부 정보

    Attributes:
        date (datetime): 조회 기준일 (KST)
        character_gender (str): 캐릭터 성별
        character_class (str): 캐릭터 직업
        character_hair (HairInfo): 캐릭터 헤어 정보 (제로인 경우 알파, 엔젤릭버스터인 경우 일반 모드)
        character_face (FaceInfo): 캐릭터 성형 정보 (제로인 경우 알파, 엔젤릭버스터인 경우 일반 모드)
        character_skin_name (str): 캐릭터 피부 명 (제로인 경우 알파, 엔젤릭버스터인 경우 일반 모드)
        additional_character_hair (HairInfo | None): 제로인 경우 베타, 엔젤릭버스터인 경우 드레스 업 모드에 적용 중인 헤어 정보
        additional_character_face (FaceInfo | None): 제로인 경우 베타, 엔젤릭버스터인 경우 드레스 업 모드에 적용 중인 성형 정보
        additional_character_skin_name (str | None): 제로인 경우 베타, 엔젤릭버스터인 경우 드레스 업 모드에 적용 중인 피부 명
    """

    date: datetime = Field(repr=False)
    character_gender: str
    character_class: str
    character_hair: HairInfo
    character_face: FaceInfo
    character_skin_name: str
    additional_character_hair: HairInfo | None
    additional_character_face: FaceInfo | None
    additional_character_skin_name: str | None



from datetime import datetime
from typing import Literal

from pydantic import BaseModel, Field, HttpUrl, field_validator

from ..types import (
    CharacterClass,
    CharacterGender,
    ItemGender,
    ItemOptionValue,
    ItemSlotName,
    PotentialGrade,
    PotentialOption,
    PresetNumber,
)


class ItemAddOption(BaseModel):
    """장비 추가 옵션

    Attributes:
        STR: STR
        DEX: DEX
        INT: INT
        LUK: LUK
        max_hp: 최대 HP
        max_mp: 최대 MP
        attack_power: 공격력
        magic_power: 마력
        armor: 방어력
        speed: 이동속도
        jump: 점프력
        boss_damage: 보스 몬스터 공격 시 데미지(%)
        damage: 데미지(%)
        all_stat: 올스탯(%)
        equipment_level_decrease: 장비 레벨 감소
    """

    str: ItemOptionValue
    int: ItemOptionValue
    dex: ItemOptionValue
    luk: ItemOptionValue
    max_hp: ItemOptionValue
    max_mp: ItemOptionValue
    attack_power: ItemOptionValue
    magic_power: ItemOptionValue
    armor: ItemOptionValue
    speed: ItemOptionValue
    jump: ItemOptionValue
    boss_damage: ItemOptionValue
    damage: ItemOptionValue
    all_stat: ItemOptionValue
    equipment_level_decrease: ItemOptionValue


class ItemBaseOption(BaseModel):
    """캐릭터 장비 기본 옵션 정보

    Attributes:
        str: STR
        dex: DEX
        int: INT
        luk: LUK
        max_hp: 최대 HP
        max_mp: 최대 MP
        attack_power: 공격력
        magic_power: 마력
        armor: 방어력
        speed: 이동속도
        jump: 점프력
        boss_damage: 보스 몬스터 공격 시 데미지(%)
        ignore_monster_armor: 몬스터방어율 무시(%)
        all_stat: 올스탯(%)
        max_hp_rate: 최대 HP(%)
        max_mp_rate: 최대 MP(%)
        base_equipment_level: 기본 착용 레벨
    """

    str: ItemOptionValue
    int: ItemOptionValue
    dex: ItemOptionValue
    luk: ItemOptionValue
    max_hp: ItemOptionValue
    max_mp: ItemOptionValue
    attack_power: ItemOptionValue
    magic_power: ItemOptionValue
    armor: ItemOptionValue
    speed: ItemOptionValue
    jump: ItemOptionValue
    boss_damage: ItemOptionValue
    ignore_monster_armor: ItemOptionValue
    all_stat: ItemOptionValue
    max_hp_rate: ItemOptionValue
    max_mp_rate: ItemOptionValue
    base_equipment_level: ItemOptionValue


class ItemEtcOption(BaseModel):
    """장비 기타 옵션 정보

    Attributes:
        str: STR
        dex: DEX
        int: INT
        luk: LUK
        max_hp: 최대 HP
        max_mp: 최대 MP
        attack_power: 공격력
        magic_power: 마력
        armor: 방어력
        speed: 이동속도
        jump: 점프력
    """

    str: ItemOptionValue
    int: ItemOptionValue
    dex: ItemOptionValue
    luk: ItemOptionValue
    max_hp: ItemOptionValue
    max_mp: ItemOptionValue
    attack_power: ItemOptionValue
    magic_power: ItemOptionValue
    armor: ItemOptionValue
    speed: ItemOptionValue
    jump: ItemOptionValue


class ItemExceptionalOption(BaseModel):
    """장비 익셉셔널 옵션 정보

    Attributes:
        STR: STR
        DEX: DEX
        INT: INT
        LUK: LUK
        max_hp: 최대 HP
        max_mp: 최대 MP
        attack_power: 공격력
        magic_power: 마력
    """

    str: ItemOptionValue
    int: ItemOptionValue
    dex: ItemOptionValue
    luk: ItemOptionValue
    max_hp: ItemOptionValue
    max_mp: ItemOptionValue
    attack_power: ItemOptionValue
    magic_power: ItemOptionValue


class ItemTotalOption(BaseModel):
    """장비 최종 옵션 정보

    Attributes:
        STR: STR
        INT: INT
        DEX: DEX
        LUK: LUK
        max_hp: 최대 HP
        max_mp: 최대 MP
        attack_power: 공격력
        magic_power: 마력
        armor: 방어력
        speed: 이동속도
        jump: 점프력
        boss_damage: 보스 몬스터 공격 시 데미지(%)
        ignore_monster_armor: 몬스터방어율 무시(%)
        all_stat: 올스탯(%)
        damage: 데미지(%)
        max_hp_rate: 최대 HP(%)
        max_mp_rate: 최대 MP(%)
        equipment_level_decrease: 장비 레벨 감소
    """

    str: ItemOptionValue
    int: ItemOptionValue
    dex: ItemOptionValue
    luk: ItemOptionValue
    max_hp: ItemOptionValue
    max_mp: ItemOptionValue
    attack_power: ItemOptionValue
    magic_power: ItemOptionValue
    armor: ItemOptionValue
    speed: ItemOptionValue
    jump: ItemOptionValue
    boss_damage: ItemOptionValue
    ignore_monster_armor: ItemOptionValue
    all_stat: ItemOptionValue
    damage: ItemOptionValue
    max_hp_rate: ItemOptionValue
    max_mp_rate: ItemOptionValue
    equipment_level_decrease: ItemOptionValue


class ItemStarforceOption(BaseModel):
    """캐릭터 장비 스타포스 옵션 정보

    Attributes:
        STR: STR
        DEX: DEX
        INT: INT
        LUK: LUK
        max_hp: 최대 HP
        max_mp: 최대 MP
        attack_power: 공격력
        magic_power: 마력
        armor: 방어력
        speed: 이동속도
        jump: 점프력
    """

    str: ItemOptionValue
    int: ItemOptionValue
    dex: ItemOptionValue
    luk: ItemOptionValue
    max_hp: ItemOptionValue
    max_mp: ItemOptionValue
    attack_power: ItemOptionValue
    magic_power: ItemOptionValue
    armor: ItemOptionValue
    speed: ItemOptionValue
    jump: ItemOptionValue


class BaseEquipmentInfo(BaseModel):
    """기본 장비 정보

    Attributes:
        item_equipment_part: 장비 부위 명
        item_equipment_slot: 장비 슬롯 위치
        item_name: 장비 명
        item_icon: 장비 아이콘
        item_description: 장비 설명
        item_shape_name: 장비 외형
        item_shape_icon: 장비 외형 아이콘
        item_gender: 장비 전용 성별
        item_total_option: 장비 최종 옵션
        item_base_option: 장비 기본 옵션
        item_add_option: 장비 추가 옵션
        item_etc_option: 장비 기타 옵션
        item_exceptional_option: 장비 익셉셔널 옵션
        equipment_level_increase: 장비 레벨 증가
        growth_exp: 성장 경험치
        growth_level: 성장 레벨
        scroll_upgrade: 업그레이드 횟수
        cuttable_count: 가위 사용 가능 횟수 (교환 불가 장비, 가위 횟수가 없는 교환 가능 장비는 255)
        golden_hammer_flag: 황금 망치 재련 적용 (1:적용, 이외 미 적용)
        scroll_resilience_count: 복구 가능 횟수
        scroll_upgradeable_count: 업그레이드 가능 횟수
        soul_name: 소울 명
        soul_option: 소울 옵션
        starforce: 스타포스 강화 단계
        starforce_scroll_flag: 놀라운 장비 강화 주문서 사용 여부 (0:미사용, 1:사용)
        item_starforce_option 장비 스타포스 옵션
        special_ring_level: 특수 반지 레벨
        date_expire: 장비 유효 기간 (KST) '2023-12-21T17:28+09:00'
    """

    part: str = Field(alias="item_equipment_part")
    slot: ItemSlotName = Field(alias="item_equipment_slot")
    name: str = Field(alias="item_name")
    icon: HttpUrl = Field(alias="item_icon")
    description: str | None = Field(alias="item_description")
    shape_name: str = Field(alias="item_shape_name")
    shape_icon: HttpUrl = Field(alias="item_shape_icon")
    gender: ItemGender = Field(alias="item_gender")
    total_option: ItemTotalOption = Field(alias="item_total_option")
    base_option: ItemBaseOption = Field(alias="item_base_option")
    add_option: ItemAddOption = Field(alias="item_add_option")
    etc_option: ItemEtcOption = Field(alias="item_etc_option")
    exceptional_option: ItemExceptionalOption = Field(alias="item_exceptional_option")
    starforce_option: ItemStarforceOption = Field(alias="item_starforce_option")
    equipment_level_increase: int
    growth_exp: int
    growth_level: int
    scroll_upgrade: int
    cuttable_count: int | None
    golden_hammer_flag: bool
    scroll_resilience_count: int
    scroll_upgradeable_count: int
    soul_name: str | None
    soul_option: str | None
    starforce: int
    starforce_scroll_flag: bool
    special_ring_level: int
    date_expire: datetime | None

    @field_validator("cuttable_count", mode="before")
    @classmethod
    def validate_cuttable_count(cls, v: str | int) -> int | None:
        v = int(v)
        return None if v == 255 else v

    @field_validator("golden_hammer_flag", mode="before")
    @classmethod
    def validate_golden_hammer_flag(cls, v: str | int) -> bool:
        match v:
            case "미적용" | "0" | 0:
                return False
            case "적용" | "1" | 1:
                return True
            case _:
                raise ValueError(f"Unexpected value on golden_hammer_flag: {v!r}")

    @field_validator("starforce_scroll_flag", mode="before")
    @classmethod
    def validate_starforce_scroll_flag(cls, v: str | int) -> bool:
        match v:
            case "미사용" | "0" | 0:
                return False
            case "사용" | "1" | 1:
                return True
            case _:
                raise ValueError(f"Unexpected value on starforce_scroll_flag: {v!r}")


class DragonEquipmentInfo(BaseEquipmentInfo):
    """에반 드래곤 장비 정보"""


class MechanicEquipmentInfo(BaseEquipmentInfo):
    """메카닉 장비 정보"""


class EquipmentInfo(BaseEquipmentInfo):
    """장비 정보

    Attributes:
        potential_grade: 잠재 옵션 등급
        potential_option_1: 잠재 옵션 첫번째 옵션
        potential_option_2: 잠재 옵션 두번째 옵션
        potential_option_3: 잠재 옵션 세번째 옵션
        additional_grade: 에디셔널 잠재 옵션 등급
        additional_option_1: 에디셔널 잠재 첫번째 옵션
        additional_option_2: 에디셔널 잠재 두번째 옵션
        additional_option_3: 에디셔널 잠재 세번째 옵션

        -- 이후는 상속된 속성
        item_equipment_part: 장비 부위 명
        item_equipment_slot: 장비 슬롯 위치
        item_name: 장비 명
        item_icon: 장비 아이콘
        item_description: 장비 설명
        item_shape_name: 장비 외형
        item_shape_icon: 장비 외형 아이콘
        item_gender: 장비 전용 성별
        item_total_option: 장비 최종 옵션
        item_base_option: 장비 기본 옵션
        equipment_level_increase: 장비 레벨 증가
        item_exceptional_option: 장비 익셉셔널 옵션
        item_add_option: 장비 추가 옵션
        growth_exp: 성장 경험치
        growth_level: 성장 레벨
        scroll_upgrade: 업그레이드 횟수
        cuttable_count: 가위 사용 가능 횟수 (교환 불가 장비, 가위 횟수가 없는 교환 가능 장비는 255)
        golden_hammer_flag: 황금 망치 재련 적용 (1:적용, 이외 미 적용)
        scroll_resilience_count: 복구 가능 횟수
        scroll_upgradeable_count: 업그레이드 가능 횟수
        soul_name: 소울 명
        soul_option: 소울 옵션
        item_etc_option: 장비 기타 옵션
        starforce: 스타포스 강화 단계
        starforce_scroll_flag: 놀라운 장비 강화 주문서 사용 여부 (0:미사용, 1:사용)
        item_starforce_option 장비 스타포스 옵션
        special_ring_level: 특수 반지 레벨
        date_expire: 장비 유효 기간 (KST)
    """

    potential_grade: PotentialGrade = Field(alias="potential_option_grade")
    potential_option_1: PotentialOption = Field(alias="potential_option_1")
    potential_option_2: PotentialOption = Field(alias="potential_option_2")
    potential_option_3: PotentialOption = Field(alias="potential_option_3")

    additional_grade: PotentialGrade = Field(alias="additional_potential_option_grade")
    additional_option_1: PotentialOption = Field(alias="additional_potential_option_1")
    additional_option_2: PotentialOption = Field(alias="additional_potential_option_2")
    additional_option_3: PotentialOption = Field(alias="additional_potential_option_3")

    @property
    def potential_options(self) -> list[PotentialOption]:
        return [
            self.potential_option_1,
            self.potential_option_2,
            self.potential_option_3,
        ]

    @property
    def additional_options(self) -> list[PotentialOption]:
        return [
            self.additional_option_1,
            self.additional_option_2,
            self.additional_option_3,
        ]


class EquipmentTitle(BaseModel):
    """캐릭터 칭호 아이템 정보

    Attributes:
        name (str): 칭호 명
        icon (HttpUrl): 칭호 아이콘
        description (str): 칭호 설명
        date_expire (datetime | None): 칭호 유효 기간
        date_option_expire (datetime | Literal["expired"] | None): 칭호 옵션 유효 기간 (expired:만료, null:무제한) (KST)
    """

    name: str = Field(alias="title_name")
    icon: HttpUrl = Field(alias="title_icon")
    description: str = Field(alias="title_description")
    date_expire: datetime | None
    date_option_expire: datetime | Literal["expired"] | None


class CharacterEquipment(BaseModel):
    """캐릭터 장비 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        character_gender (CharacterGender): 캐릭터 성별
        character_class (CharacterClass): 캐릭터 직업
        items (list[EquipmentInfo]): 장비 정보
        dragon_items (list[DragonEquipmentInfo]): 에반 드래곤 장비 정보 (에반인 경우 응답)
        mechanic_items (list[MechanicEquipmentInfo]): 메카닉 장비 정보 (메카닉인 경우 응답)
        title (EquipmentTitle | None): 칭호 정보
        preset_no (PresetNumber): 적용 중인 어빌리티 프리셋 번호
        preset1 (list[EquipmentInfo]): 1번 프리셋 장비 정보
        preset2 (list[EquipmentInfo]): 2번 프리셋 장비 정보
        preset3 (list[EquipmentInfo]): 3번 프리셋 장비 정보
    """

    date: datetime = Field(repr=False)
    character_gender: CharacterGender
    character_class: CharacterClass
    items: list[EquipmentInfo] = Field(alias="item_equipment")
    dragon_items: list[DragonEquipmentInfo] = Field(alias="dragon_equipment")
    mechanic_items: list[MechanicEquipmentInfo] = Field(alias="mechanic_equipment")
    title: EquipmentTitle | None
    preset_no: PresetNumber
    preset1: list[EquipmentInfo] = Field(alias="item_equipment_preset1")
    preset2: list[EquipmentInfo] = Field(alias="item_equipment_preset2")
    preset3: list[EquipmentInfo] = Field(alias="item_equipment_preset3")



from pydantic import BaseModel, Field


class OverallRankingInfo(BaseModel):
    """종합 랭킹 상세 정보

    Attributes:
        date: 랭킹 업데이트 일자 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        ranking: 종합 랭킹 순위
        character_name: 캐릭터 명
        world_name: 월드 명
        class_name: 직업 명
        sub_class_name: 전직 직업 명
        character_level: 캐릭터 레벨
        character_exp: 캐릭터 경험치
        character_popularity: 캐릭터 인기도
        character_guildname: 길드 명
    """

    date: str
    ranking: int
    character_name: str
    world_name: str
    class_name: str  # FIXME: RankingCharacterClass
    sub_class_name: str  # FIXME: RankingCharacterSubClass
    character_level: int
    character_exp: int
    popularity: int = Field(alias="character_popularity")
    guild_name: str | None = Field(alias="character_guildname")

    @property
    def 레벨(self) -> int:
        return self.character_level

    @property
    def 경험치(self) -> int:
        return self.character_exp

    @property
    def 인기도(self) -> int:
        return self.popularity

    @property
    def 길드명(self) -> str | None:
        return self.guild_name


class OverallRanking(BaseModel):
    """종합 랭킹 정보

    Attributes:
        ranking: 종합 랭킹 정보
    """

    ranking: list[OverallRankingInfo]



"""길드 정보 조회 API를 제공하는 모듈입니다.

Note:
    - 2023년 12월 21일부터 데이터를 조회할 수 있습니다.
    - 캐릭터 정보 조회 API는 일자별 데이터로 매일 오전 1시부터 전일 데이터 조회가 가능합니다.
      (예를 들어, 12월 22일 데이터를 조회하면 22일 00시부터 23일의 00시 사이의 데이터가 조회됩니다.)
    - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
"""

from datetime import datetime

import maplestory.utils.date as dates
import maplestory.utils.kst as kst
from maplestory.models.guild import GuildBasic, GuildModel
from maplestory.models.types import WorldName
from maplestory.utils.network import fetch


def get_guild_id(guild_name: str, world_name: WorldName) -> str:
    """길드 식별자(gcid) 정보를 조회합니다.

    Args:
        guild_name (str): 길드명.
        world_name (str): 월드명.
            Available values: 스카니아, 베라, 루나, 제니스, 크로아, 유니온, 엘리시움,
                이노시스, 레드, 오로라, 아케인, 노바, 리부트, 리부트2, 버닝, 버닝2, 버닝3

    Returns:
        str: 길드 식별자(gcid) 정보.
    """

    path = "/maplestory/v1/guild/id"
    query = {
        "guild_name": guild_name,
        "world_name": world_name,
    }
    response = fetch(path, query)

    return GuildModel.model_validate(response).id


def get_basic_info_by_id(
    guild_id: str,
    date: datetime = kst.yesterday(),
) -> GuildBasic:
    """길드 기본 정보를 조회합니다.

    Args:
        guild_id (str): 길드 식별자.
        date (datetime): 조회 기준일(KST).

    Returns:
        GuildBasic: 길드의 기본 정보.
    """

    path = "/maplestory/v1/guild/basic"
    query = {
        "oguild_id": guild_id,
        "date": dates.to_string(date),
    }
    response = fetch(path, query)

    return GuildBasic.model_validate(response)



from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field


class UnionRaiderBlockPosition(BaseModel):
    """블록이 차지하고 있는 영역 좌표들

    Attributes:
        x: 블록 X좌표
        y: 블록 Y좌표
    """

    x: int
    y: int


class UnionRaiderBlockControlPoint(BaseModel):
    """블록 기준점 좌표

    Attributes:
        x: 블록 기준점 X좌표
        y: 블록 기준점 Y좌표

    Notes:
        - 중앙 4칸 중 오른쪽 아래 칸이 x=0, y=0 포지션
        - 좌측으로 1칸씩 이동하면 x가 1씩 감소
        - 우측으로 1칸씩 이동하면 x가 1씩 증가
        - 아래로 1칸씩 이동하면 y가 1씩 감소
        - 위로 1칸씩 이동하면 y가 1씩 증가
    """

    x: int
    y: int


class UnionRaiderBlock(BaseModel):
    """유니온 블록 배치 정보

    Attributes:
        type: 블록 모양 (전사, 마법사, 궁수, 도적, 해적, 메이플M, 하이브리드)
        character_class: 블록 해당 캐릭터 직업
        level: 블록 해당 캐릭터 레벨
        control_point: 블록 기준점 좌표
        position: 블록이 차지하고 있는 영역 좌표들
    """

    type: str = Field(alias="block_type")
    character_class: str = Field(alias="block_class")
    level: str = Field(alias="block_level")
    control_point: UnionRaiderBlockControlPoint = Field(alias="block_control_point")
    position: list[UnionRaiderBlockPosition] = Field(alias="block_position")


class UnionRaiderInnerStat(BaseModel):
    """유니온 공격대 안쪽 지역 효과 배치 정보

    Attributes:
        stat_field_id: 공격대 배치 위치 (11시 방향부터 시계 방향 순서대로 0~7)
        stat_field_effect: 해당 지역 점령 효과

    Examples:
        {
            "stat_field_id": "0",
            "stat_field_effect": "유니온 STR"
        },
        {
            "stat_field_id": "4",
            "stat_field_effect": "유니온 공격력"
        },
        {
            "stat_field_id": "6",
            "stat_field_effect": "유니온 최대 HP"
        }
    """

    stat_field_id: str
    stat_field_effect: str


class UnionRaider(BaseModel):
    """유니온 공격대 정보

    Attributes:
        date: 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        raider_stats: 유니온 공격대원 효과
        occupied_stats: 유니온 공격대 점령 효과
        inner_stats: 유니온 공격대 안쪽 지역 효과 정보
        blocks: 유니온 블록 배치 정보
    """

    date: datetime = Field(repr=False)
    raider_stats: list[str] = Field(alias="union_raider_stat")
    occupied_stats: list[str] = Field(alias="union_occupied_stat")
    inner_stats: list[UnionRaiderInnerStat] = Field(alias="union_inner_stat")
    blocks: list[UnionRaiderBlock] = Field(alias="union_block")

    def model_post_init(self, __context: Any) -> None:
        self.raider_stats = sorted(self.raider_stats)
        self.occupied_stats = sorted(self.occupied_stats)

    @property
    def 공격대원_효과(self) -> list[str]:
        return self.raider_stats

    @property
    def 공격대_점령_효과(self) -> list[str]:
        return self.occupied_stats


# (.+) (\d+%?) (증가|감소|회복)
# (.+) (\d+%?) (.+)

# 경험치 획득량 10% 증가
# 공격 시 20%의 확률로 데미지 20% 증가
# 공격력/마력 20 증가
# 메소 획득량 4% 증가
# 방어율 무시 6% 증가
# 버프 지속시간 25% 증가
# 보스 몬스터 공격 시 데미지 6% 증가
# 상태 이상 내성 4 증가
# 소환수 지속시간 10% 증가
# 스킬 재사용 대기시간 6% 감소
# 적 공격마다 70%의 확률로 순수 HP의 8% 회복
# 최대 HP 2000 증가
# 최대 HP 5% 증가
# 크리티컬 데미지 6% 증가
# 크리티컬 확률 4% 증가
# DEX 80 증가
# INT 80 증가
# LUK 80 증가
# STR 100 증가
# STR 80 증가
# STR, DEX, LUK 40 증가

# "union_raider_stat": [
#         "INT 80 증가",
#         "보스 몬스터 공격 시 데미지 6% 증가",
#         "최대 HP 5% 증가",
#         "INT 80 증가",
#         "소환수 지속시간 10% 증가",
#         "STR 80 증가",
#         "INT 80 증가",
#         "크리티컬 확률 4% 증가",
#         "적 공격마다 70%의 확률로 순수 HP의 8% 회복",
#         "공격 시 20%의 확률로 데미지 20% 증가",
#         "STR 80 증가",
#         "크리티컬 데미지 6% 증가",
#         "DEX 80 증가",
#         "경험치 획득량 10% 증가",
#         "DEX 80 증가",
#         "최대 HP 2000 증가",
#         "STR 80 증가",
#         "상태 이상 내성 4 증가",
#         "INT 80 증가",
#         "STR 80 증가",
#         "STR, DEX, LUK 40 증가",
#         "최대 HP 2000 증가",
#         "INT 80 증가",
#         "STR 100 증가",
#         "STR 100 증가",
#         "STR 100 증가",
#         "LUK 80 증가",
#         "공격력/마력 20 증가",
#         "LUK 80 증가",
#         "DEX 80 증가",
#         "크리티컬 확률 4% 증가",
#         "STR 80 증가",
#         "버프 지속시간 25% 증가",
#         "방어율 무시 6% 증가",
#         "DEX 80 증가",
#         "메소 획득량 4% 증가",
#         "스킬 재사용 대기시간 6% 감소",
#         "DEX 80 증가",
#         "LUK 80 증가"
#     ],
#     "union_occupied_stat": [
#         "STR 70 증가",
#         "크리티컬 데미지 20.00% 증가",
#         "크리티컬 확률 20% 증가",
#         "방어율 무시 40% 증가",
#         "최대 MP 250 증가",
#         "공격력 9 증가",
#         "보스 몬스터 공격 시 데미지 40% 증가",
#         "마력 1 증가"
#     ],



from enum import Enum

import maplestory.utils.kst as kst


class BaseEnum(Enum):
    @classmethod
    def keys(cls):
        return list(cls.__members__)

    @classmethod
    def values(cls):
        return [c.value for c in cls]


class WorldType(BaseEnum):
    일반 = 0  # normal
    리부트 = 1  # reboot

    # NORMAL = 0 # 일반
    # REBOOT = 1 # 리부트


class GuildRankType(BaseEnum):
    주간명성치 = 0  # Weekly Fame
    플레그레이스 = 1  # Flag Race
    지하수로 = 2  # Underground Waterway

    # WEEKLY_FAME = 0  # 주간 명성치
    # FLAG_RACE = 1  # 플래그 레이스
    # UNDERGROUND_WATERWAY = 2  # 지하 수로


class MinimumDate(BaseEnum):
    # CHARACTER = kst.datetime(2023, 12, 21)
    # RANK = kst.datetime(2023, 12, 22)

    # CUBE = kst.datetime(2022, 11, 25)
    # STARFORCE = kst.datetime(2023, 12, 27)
    # POTENTIAL = kst.datetime(2024, 1, 25)

    캐릭터 = kst.datetime(2023, 12, 21)
    유니온 = kst.datetime(2023, 12, 21)
    길드 = kst.datetime(2023, 12, 21)
    랭킹 = kst.datetime(2023, 12, 22)
    큐브 = kst.datetime(2022, 11, 25)
    스타포스 = kst.datetime(2023, 12, 27)
    잠재능력 = kst.datetime(2024, 1, 25)


# from enum import Enum
# from datetime import datetime
# from dataclasses import dataclass

# @dataclass
# class MinimumDate(Enum):
#     CHARACTER = CHARACTER_MINIMUM_DATE
#     GUILD = GUILD_MINIMUM_DATE
#     UNION = UNION_MINIMUM_DATE
#     RANK = RANK_MINIMUM_DATE
#     CUBE = CUBE_MINIMUM_DATE
#     STARFORCE = STARFORCE_MINIMUM_DATE
#     POTENTIAL = POTENTIAL_MINIMUM_DATE

# def is_available_date(date: kst.AwareDatetime, category: MinimumDate) -> bool:
#     kst.validate(date)

#     return match category:
#         case MinimumDate.CHARACTER:
#             return date >= MinimumDate.CHARACTER.value
#         case MinimumDate.GUILD:
#             return date >= MinimumDate.GUILD.value
#         case MinimumDate.UNION:
#             return date >= MinimumDate.UNION.value
#         case MinimumDate.RANK:
#             return date >= MinimumDate.RANK.value
#         case MinimumDate.CUBE:
#             return date >= MinimumDate.CUBE.value
#         case MinimumDate.STARFORCE:
#             return date >= MinimumDate.STARFORCE.value
#         case MinimumDate.POTENTIAL:
#             return date >= MinimumDate.POTENTIAL.value


# class TestEnum(Enum):
#     NORMAL = '0' # 일반
#     REBOOT = '1' # 리부트

# print(f"{TestEnum.NORMAL == '0' = }")


# class DateEnum(datetime, ReprEnum):
#     """
#     Enum where members are also (and must be) ints
#     """

# class TestDateEnum(DateEnum):
#     NORMAL = datetime(2023, 12, 21) # 일반

# print(f"{TestDateEnum.NORMAL == datetime(2023, 12, 21) = }")
# # TypeError: 'datetime.datetime' object cannot be interpreted as an integer

# class TestIntEnum(IntEnum):
#     NORMAL = 0 # 일반
#     REBOOT = 1 # 리부트

# print(f"{TestIntEnum.NORMAL == 0 = }")


# class TestStrEnum(StrEnum):
#     NORMAL = "0" # 일반
#     REBOOT = "1" # 리부트

# print(f"{TestStrEnum.NORMAL == "0" = }")



"""길드 정보 조회 API를 제공하는 모듈입니다.

Note:
    - 2023년 12월 21일부터 데이터를 조회할 수 있습니다.
    - 캐릭터 정보 조회 API는 일자별 데이터로 매일 오전 1시부터 전일 데이터 조회가 가능합니다.
      (예를 들어, 12월 22일 데이터를 조회하면 22일 00시부터 23일의 00시 사이의 데이터가 조회됩니다.)
    - 게임 콘텐츠 변경으로 ocid가 변경될 수 있습니다. ocid 기반 서비스 갱신 시 유의해 주시길 바랍니다.
"""

from datetime import datetime

import maplestory.utils.date as dates
import maplestory.utils.kst as kst
from maplestory.models.guild import GuildBasic, GuildModel
from maplestory.models.types import WorldName
from maplestory.utils.network import fetch


def get_id(guild_name: str, world_name: WorldName) -> str:
    """길드 식별자(gcid) 정보를 조회합니다. Fetches the guild identifier (gcid).

    Args:
        guild_name : 길드명. The name of the guild.
        world_name : 월드명. The name of the world.

    Returns:
        str: 길드 식별자(gcid) 정보. The guild identifier (gcid).
    """

    path = "/maplestory/v1/guild/id"
    query = {
        "guild_name": guild_name,
        "world_name": world_name,
    }
    response = fetch(path, query)

    return GuildModel.model_validate(response).id


def get_basic_info_by_id(
    guild_id: str,
    date: datetime = kst.yesterday(),
) -> GuildBasic:
    """길드 기본 정보를 조회합니다. Fetches the basic information of the guild.

    Args:
        guild_id : 길드 식별자. The identifier of the guild.
        date : 조회 기준일(KST). Reference date for the query (KST).

    Returns:
        GuildBasic: 길드의 기본 정보. The basic information of the guild.
    """

    path = "/maplestory/v1/guild/basic"
    query = {
        "oguild_id": guild_id,
        "date": dates.to_string(date),
    }
    response = fetch(path, query)

    return GuildBasic.model_validate(response)



from datetime import datetime

from pydantic import BaseModel, Field


class HexaMatrixLinkedSkill(BaseModel):
    """연결된 스킬 정보

    Attributes:
        hexa_skill_id (str): HEXA 스킬 ID
    """

    hexa_skill_id: str


class HexaMatrixCore(BaseModel):
    """HEXA 코어 상세 정보

    Attributes:
        hexa_core_name (str): 코어 명
        hexa_core_level (int): 코어 레벨
        hexa_core_type (str): 코어 타입
        linked_skill (list[HexaMatrixLinkedSkill]): 연결된 스킬
    """

    hexa_core_name: str
    hexa_core_level: int
    hexa_core_type: str
    linked_skill: list[HexaMatrixLinkedSkill]


class HexaMatrix(BaseModel):
    """HEXA 코어 정보

    Attributes:
        date (datetime): 조회 기준일 (KST, 일 단위 데이터로 시, 분은 일괄 0으로 표기)
        character_hexa_core_equipment (list[HexaMatrixCore] | None): HEXA 코어 정보

    Notes:
        HEXA 코어 정보가 없는 경우 None으로 표기됩니다.
    """

    date: datetime = Field(repr=False)
    character_hexa_core_equipment: list[HexaMatrixCore] | None
